from .abstract_apis import APIBridge,ApiGUI
from .abstract_rpcs import RPCBridge,RPCGUIManager
from .abstract_blockchain_functions import *
from abstract_utilities import safe_dump_to_file,create_and_read_json
class abstractBlockchainManager:
    def __init__(self,rpc_mgr=None,api_mgr=None,abi_mgr=None,w3_mgr=None):
        self.w3_mgr=w3_mgr or w3Manager()
        self.script_path = os.path.abspath(__file__)
        self.directory_path = os.path.dirname(self.script_path)
        self.data_path = makeAllDirs(os.path.join(self.directory_path,'data'))
        self.manager_key_path= makeAllDirs(os.path.join(self.data_path,'manager_keys.json'))
        os.path.dirname(os.path.abspath(__file__))
        os.makedirs(os.path.join(get_directory_path(), 'data'),exist_ok=True)
        os.makedirs(os.path.join(get_data_directory(), 'source_codes'),exist_ok=True)
        self.address_manager=create_and_read_json(file_path=self.manager_key_path,data={})
        self.rpc_mgr = rpc_mgr
        if self.rpc_mgr == None:
            self.rpc_mgr=RPCBridge()
        self.api_mgr = api_mgr
        if self.api_mgr == None:
            self.api_mgr=APIBridge()
        self.abi_mgr=abi_mgr
        if self.abi_mgr == None:
            self.abi_mgr=ABIBridge(w3_mgr=self.w3_mgr)
    def derive_address_values(self,address,rpc_js=None):
        source_code_directory = get_source_code_path(address)
        source_code_path = os.path.join(source_code_directory,'source_code.json')
        dict_info_path = os.path.join(source_code_directory,'dict_info.json')
        rpc_path = os.path.join(source_code_directory,'rpc_data.json')
        source_code = read_json(file_path=source_code_path)
        rpc_js = read_json(file_path=rpc_path)
        self.w3_mgr.create_w3(address,rpc=rpc_js)
        abi=source_code[0]['ABI']
        self.get_bridge(address,abi=abi)
        normalized_address = get_normalized_address(address)
        contract_info={"normalized_address":normalized_address,
                                                  "checksum_address":address,
                                                  "rpc_js":rpc_js,
                                                  "dict_info_path":dict_info_path,
                                                  "rpc_path":rpc_js,
                                                  "source_code_path":source_code_path,
                                                  "source_code_directory":source_code_directory,
                                                  "function_calls":{},"abi":abi}
        self.update_contract_info(normalized_address,contract_info)
    def get_contract_info(self,address):
        normalized_address = get_normalized_address(address=address)
        if normalized_address not in self.address_manager:
            self.derive_address_values(address)
        return self.address_manager[normalized_address]
    def get_bridge(self,address):
        contract_info = self.get_contract_info(address)
        return self.abi_mgr.create_abi_bridge(address,abi=contract_info['abi'],rpc_js=contract_info['rpc_js'])
    def update_contract_info(self,address,contract_info):
        normalized_address = get_normalized_address(address=address)
        old_contract_info={}
        for key,value in contract_info.items():
            old_contract_info[key] = value
        self.address_manager[normalized_address]=old_contract_info
        safe_dump_to_file(file_path = self.address_manager[normalized_address]["dict_info_path"],data=self.address_manager[normalized_address])
        safe_dump_to_file(file_path=self.manager_key_path,data=self.address_manager)
    def change_rpc_selection(self,address):
         contract_info = self.get_contract_info(address)
         rpc_gui_mgr = RPCGUIManager(rpc_js=contract_info['rpc_js'])
         contract_info['rpc_js'] = rpc_gui_mgr.rpc_values
         self.update_contract_info(address,contract_info)
         self.rpc_mgr.rpc_js = contract_info['rpc_js']
         self.w3_mgr.create_w3(address, rpc=self.rpc_mgr.rpc_js)
         self.abi_mgr.update_abi_bridge(address,rpc_js=self.rpc_mgr.rpc_js)
    def call_function(self,address,function_name,*args):
        contract_info = self.get_contract_info(address)
        if function_name not in contract_info:
            contract_info['function_calls'][function_name] = []
        response = self.abi_mgr.call_function(address,function_name,*args) 
        contract_info['function_calls'][function_name].append({'args':args,'response':response})
        self.update_contract_info(address,contract_info)
        return response



from abstract_gui import get_push,AbstractWindowManager
from .abstract_rpcs import rpc_win_while,get_rpc_layout,get_rpc_js,RPCBridge
from .abstract_abis import ABIBridge
from .abstract_accounts import ACCTBridge
import codecs
def mkBool(x):
    if str(x).lower() in ['true','t','yes','y','0']:
        return str(0)
    return str(1)
def isUintSt(x):
    if x[:len('uint')].lower() == 'uint' and str(x).lower() in ['uint8','uint16','uint32','uint64','uint128','uint256']:
        return True
    return False
def isIntSt(x):
    if x[:len('int')].lower() == 'int' and str(x).lower() in ['int8','int16','int32','int64','int128','int256']:
        return True
    return False
def isAddress(x):
    if x[:len('address')].lower() == 'address':
        return True
def isBool(x):
    if x[:len('bool')].lower() == 'bool':
        return True
def isString(x):
    if x[:len('string')].lower() == 'string':
        return True
def isBytes(x):
    if x[:len('bytes')].lower() == 'bytes':
        return True
def isTuple(x):
    if x[:len('tuple')].lower() == 'tuple':
        return True
def getCodex(obj: str) -> str:
    # Convert a hex string to bytes and then decode to UTF-8
    b = bytes.fromhex(obj)
    return codecs.decode(b, 'UTF-8')
def get_keccak(obj: str) -> str:
    # Create a keccak hash of a string and get its hexadecimal representation
    return get_w3().keccak(text=str(obj)).hex()
def read_hex(hb: bytes) -> str:
    # Convert bytes to hex string
    return hb.hex()
def get_hex_data(obj: str) -> int:
    # Convert a hex string to an integer
    return int(obj, 16)
def can_be_converted_from_keccak(obj: str) -> bool:
    try:
        # Try to get the keccak hash of the object
        keccak_hash = get_keccak(obj)
        # Check if the object can be decoded from hex to UTF-8
        decoded_string = getCodex(keccak_hash)
        return True
    except Exception:
        return False
def string_to_bytes(input_type: str, value: (str or bytes)) -> str:
    # If the type is dynamic "bytes"
    if input_type == "bytes":
        if isinstance(value, str) and not value.startswith("0x"):
            value_bytes = value.encode('utf-8')
        elif isinstance(value, str) and value.startswith("0x"):
            value_bytes = bytes.fromhex(value[2:])
        else:
            value_bytes = value
    else:
        # Determine the target size from the input_type
        size = int(input_type[5:])
        
        # If the value is a regular string, encode it
        if isinstance(value, str) and not value.startswith("0x"):
            value_bytes = value.encode('utf-8')
        # If the value starts with "0x", convert it to bytes
        elif isinstance(value, str) and value.startswith("0x"):
            value_bytes = bytes.fromhex(value[2:])
        # If the value is already bytes, just use it
        else:
            value_bytes = value
        
        # Adjust size of the bytes value
        if len(value_bytes) > size:
            # Truncate if too long
            value_bytes = value_bytes[:size]
        elif len(value_bytes) < size:
            # Pad with zeros if too short
            value_bytes += b'\x00' * (size - len(value_bytes))
        
    # Return as hex string
    return '0x' + value_bytes.hex()
def get_type(input_type, value, output=False):
    if value == '':
        return str('')
    if isUintSt(input_type) or isIntSt(input_type):
        return int(value)
    elif isBytes(input_type):
        if output == True:
            if can_be_converted_from_keccak(value):
                # Try to get the keccak hash of the object
                keccak_hash = get_keccak(obj=value)
                # Check if the object can be decoded from hex to UTF-8
                decoded_string = getCodex(obj=keccak_hash)
                return str(decoded_string)
        return str(string_to_bytes(input_type=input_type,value=value))
    elif isTuple(input_type):
        return tuple(value)
    elif isAddress(input_type):
        return try_check_sum(str(value))
    elif isBool(input_type):
        value = bool(str(value))
        if output == True:
            return {'0':"True","1":"False"}[str(value)]
        return value
    return str(value)
            
def get_w3():
    if "rpc_manager" not in new_bridge_global:
        return RPCBridge().w3
    if not isinstance(new_bridge_global["rpc_manager"],type(RPCBridge())):
        print("rpc w3 no good")
        return RPCBridge().w3
    return new_bridge_global["rpc_manager"].w3
def get_size(input_type):
    if isUintSt(input_type) or isIntSt(input_type) or isString(input_type):
        return 24
    elif isBytes(input_type):
        return len("0x****************************************************************")+4
    elif isAddress(input_type):
        return len("0x****************************************")+4
    elif isBool(input_type):
        return 4
    else:
        return len("0x****************************************************************")+4
    
def try_check_sum(address:str):
    try:
        address = get_w3().to_checksum_address(str(address))
        return address
    except:
        return False
def lsNum():
    return str('1,2,3,4,5,6,7,8,9,0').split(',')
def isNum(x):
    if isInt(x):
        return True
    for k in range(0,len(str(x))):
        if str(x)[k] not in lsNum():
            return False
    return True
def isLs(ls):
    if type(ls) is list:
        return True
    return False
def isList(x):
    if x[-1] == ']':
        if x[:-1].split('[')[-1] == '':
            return [x],'*',0
        lsN = []
        for i in range(0,int(x[:-1].split('[')[-1])):
            lsN.append(x)    
        return lsN,i,0
    return x,1,0
def isInt(x):
  if type(x) is int:
    return True
  return False
def call_function(function_name):
    values = new_window_mgr.get_values()
    inputs = {}
    output_keys= []
    for key, value in values.items():
        if f"-INPUT_{function_name}_" in key:
            input_name = f"{key.split('_')[-2]}"
            input_type = f"{key.split('_')[-1][:-1]}"# Assuming the name of the input is second to the last in the split result
            inputs[input_name] = get_type(input_type,value)
        elif f"-OUTPUT_{function_name}_" in key:
            if key not in output_keys:
                output_keys.append(key)
    try:
        # If there's only one input and it's of type address
        #if len(list(inputs.keys())) == 1:
        #    result = new_bridge_global["abi_manager"].call_function(inputs[list(inputs.keys())[0]],function_name=function_name)
        # For multiple inputs, unpack them as positional arguments
        if inputs:
            args = tuple(inputs.values())  # Convert the dictionary values to a tuple
            contract_bridge = new_bridge_global["abi_manager"].create_functions(*args,function_name=function_name,subsinstance="functions")
        else:
            contract_bridge = new_bridge_global["abi_manager"].create_functions(function_name=function_name,subsinstance="functions")
        if get_funciton_mutability(function_name).lower() in ["pure","view"]:
            result = contract_bridge.call()
        else:
            txn_info = new_bridge_global["account_manager"].build_txn(contract_bridge=contract_bridge)
            result = new_bridge_global["account_manager"].send_transaction(tx_info=txn_info)
        if len(output_keys)>0:
            if len(output_keys) == 1:
                new_window_mgr.update_values(key=output_keys[0], args={"value": str(get_type(output_keys[0].split('_')[-1][:-1],result,output=True))})
            if len(output_keys)>1:
                if not isinstance(result,list):
                    if isinstance(result,str):
                        result=result.split(',')
                    if isinstance(result,set):
                        result=list(result)
                    else:   
                        result = [result]
                for each_output_key in output_keys:
                    num = int(each_output_key[len(f"-OUTPUT_{function_name}_"):].split('_')[0])
                    new_window_mgr.update_values(key=each_output_key, args={"value": str(get_type(each_output_key.split('_')[-1][:-1],result[num],output=True))})
    except Exception as e:
        print(f"Error calling function: {e} \nInputs: {inputs}")
def get_funciton_mutability(function_name):
    for each in new_bridge_global["abi_manager"].abi:
        if each["type"] == "function":
            if each["name"] == function_name:
                return each["stateMutability"]
def contract_win_while(event: str):
    if event == "-CALL_ALL_READ_ONLY-":
        read_only_list = new_bridge_global["abi_manager"].get_read_only_functions()
        for each in read_only_list:
            call_function(each)
    if "-CALL-" in event:
        # Extracting inputs and outputs associated with the function
        call_function(event.split("-")[2])
def win_while(event: str):
    rpc_win_while(event)
    values = new_window_mgr.get_values()
    rpc_manager = get_rpc()
    if event == "-OK_RPC-":
        new_bridge_global["rpc_manager"]=rpc_manager
        print(new_bridge_global["rpc_manager"].rpc_js)
    if event == "-GET_ABI-":
        contract_address = values["-CONTRACT_ADDRESS-"]
        new_bridge_global["abi_manager"] = ABIBridge(contract_address=contract_address, rpc=rpc_manager.rpc_js)
        function_js = {}
        for each in new_bridge_global["abi_manager"].abi:
            if each["type"] == "function":
                if each["stateMutability"] not in function_js:
                    function_js[each["stateMutability"]] = []
                inputs = _parse_io(each.get("inputs", []),function_name=each["name"])
                outputs = _parse_io(each.get("outputs", []), is_output=True,function_name=each["name"])
                layout = inputs + outputs  # Combine inputs and outputs
                button = [agf("Button", args={"button_text": f"Call {each['name']}", "key": f"-CALL-{each['name']}-"})]
                layout.append(button)
                function_js[each["stateMutability"]].append(
                    agf("Frame", args={"title": each["name"], "layout": layout})
                )
        # Organizing framed groups with 4 columns per row
        all_layouts = [agf("Button", args={"button_text": f"Call All Read Only", "key": "-CALL_ALL_READ_ONLY-"})]
        for state, funcs in function_js.items():
            rows = [funcs[i:i+7] for i in range(0, len(funcs), 7)]
            state_layout = []
            for row in rows:
                state_layout.append(row)
            all_layouts.append([agf("Frame", args={"title": state, "layout": state_layout})])
        new_window = new_window_mgr.get_new_window(title=f"{new_window_mgr.get_values(window=new_bridge_global['main_window'])['-NETWORK_NAME-']} contract {contract_address}", layout=all_layouts, event_function="contract_win_while")
        new_window_mgr.while_basic(window=new_window)
    if event == "-DERIVE_RPC-":
        contract_address = values["-CONTRACT_ADDRESS-"]
        new_bridge_global["abi_manager"]=determine_correct_rpc(contract_address=contract_address,rpc_list=new_bridge_global["rpc_list"])
        rpc = new_bridge_global["abi_manager"].rpc_manager.rpc_js
        if isinstance(rpc,dict):
            for key,value in rpc.items():
                new_window_mgr.update_values(key=get_rpc_js()[key],args={"value":value})
    if event == "-ASSOCIATE_ACCOUNT-":
        new_bridge_global["account_manager"] = ACCTBridge(env_key=values["-ACCOUNT_ENV_KEY-"],rpc=get_rpc())
        new_window_mgr.update_values(key="-ACCOUNT_ADDRESS-",args={"value":new_bridge_global["account_manager"].account_address})
def _parse_io(io_data,function_name:str, is_output=False):
    layout = []
    for i, io_type in enumerate(io_data):
        text = f"{io_type['name']}({io_type['type']}): "
        key_suffix = "OUTPUT" if is_output else "INPUT"
        key = f"-{key_suffix}_{function_name}_{i}_{io_type['name']}_{io_type['type']}-"
        component_type="Input"
        size = (get_size(io_type['type']), 1)
        if io_type['type'].lower() in ["bytes","string"]:
            component_type = "Multiline"
            size = (get_size(io_type['type']), 3)
        if is_output:
            layout.append([agf("Text", args={"text": text}),get_push(), agf(component_type, args={"size": size,"key": key, "disabled": True})])
        else:
            layout.append([agf("Text", args={"text": text}),get_push(),  agf(component_type, args={"size": size, "key": key})])
    return layout
def get_account_layout():
    frame_layout = [[agf("Text", args={"text": "ENV Key:"}),get_push(),agf("Input",args={"key":"-ACCOUNT_ENV_KEY-"})],
                    [agf("Text", args={"text": "Account Address"}),get_push(),agf("Input",args={"default_text":"No Address Found","key":"-ACCOUNT_ADDRESS-","disabled":True})],
                    [agf("Button", args={"button_text":"Associate Account","enable_events":True,"key":"-ASSOCIATE_ACCOUNT-"})]]
    return [agf("Frame", "Account", args={"layout":frame_layout})]
# If you need the ABI helper function in the new script, define it here 
def get_abi():
    frame_layout = [agf("Input",args={"key":"-CONTRACT_ADDRESS-"}),
                    agf("Button", args={"button_text":"GET ABI","enable_events":True,"key":"-GET_ABI-"}),
                    agf("Button", args={"button_text":"DERIVE RPC","enable_events":True,"key":"-DERIVE_RPC-"})]
    return agf("Frame", "ABI", frame_layout)
def get_rpc():
    rpc={}
    rpc_js = get_rpc_js()
    for rpc_key,window_key in rpc_js.items():
        rpc[rpc_key] = new_window_mgr.get_values(window=new_bridge_global["main_window"])[window_key]
    return RPCBridge(rpc)
def determine_correct_rpc(contract_address:str,rpc_list:list=None):
    if rpc_list == None:
        if "rpc_list" in new_bridge_global:
            if new_bridge_global["rpc_list"] != None:
                rpc_list = new_bridge_global["rpc_list"]
            else:
                rpc_list = RPCBridge().get_default_rpc_list()
        else:
            rpc_list = RPCBridge().get_default_rpc_list()
    for rpc in rpc_list:
        abi_manager = ABIBridge(contract_address=contract_address,rpc=rpc)
        if isinstance(abi_manager.abi,list):
            break
    return abi_manager
def abstract_contract_console_main(rpc_list:list=None):
    if rpc_list == None:
        rpc_list = RPCBridge().get_default_rpc_list()
    new_bridge_global["rpc_list"]=rpc_list
    # Get the rpc_layout and other associated values
    rpc_layout= [[agf("Frame", "RPC_LAY",args={"layout":get_rpc_layout(rpc_list=new_bridge_global["rpc_list"],window_mgr=new_window_mgr)})]]
    # Construct the final layout
    new_layout = [[get_account_layout()],get_abi(),rpc_layout]
    # Create and run the window
    window_mgr = AbstractWindowManager()
    new_window = window_mgr.add_window(title="New Blockchain Console", layout=[new_layout], event_function=win_while)
    new_window_mgr.while_window(window=new_window)
abstract_contract_console_main()



import json
import os
def reader(file):
    with open(file, 'r') as f:
        text = f.read()
        return text
def pen(paper, place):
    with open(place, 'w') as f:
        f.write(str(paper))
        f.close()
def changeGlob(x,v):
    globals()[x] = v
def readLinesSpec(x):
    lsN = []
    lines = x.split('\n')
    for i in range(0,len(lines)):
        lines[i] = eatOuterMod(lines[i],['\n','\t',' ','']).replace('\n','')
        if lines[i] not in ['\n',' \n','',' ','\t']  and len(str(eatAllMod(lines[i],['\n','\t',' ','/',';','']))) != 0:
            lsN.append(lines[i]+'\n')
    return lsN
def lsChangeGlob(x,y):
    for i in range(0,len(x)):
        changeGlob(x[i],y[i])

def adSt(x,y):
    return str(x)+str(y)
def lsCheck(x,ls):
    for i in range(0,len(ls)):
        if x == ls[i]:
            return True
    return False
def currIs(x,ls):
    for i in range(0,len(ls)):
        if x[:len(ls[i])] == ls[i]:
            return x[len(ls[i]):]
    return False
def currIsNeg(x,ls):
    for i in range(0,len(ls)):
        if x[len(x)-len(ls[i]):] == ls[i]:
            return x[:len(x)-len(ls[i])] 
    return False
def eatInner(x,ls):
    z = x
    for i in range(0,len(x)):
        if currIs(z,ls) == False:
            return z
        else:
            z = currIs(z,ls)
    return z
def eatOuter(x,ls):
    z = x
    for i in range(0,len(x)):
        if currIsNeg(z,ls) == False:
            input(z)
            return z
        else:
            z = currIsNeg(z,ls)

    return z
def eatAll(x,ls):
    y = eatInner(x,ls)
    z = eatOuter(y,ls)
    return z
def countIt(x,y):
    if y in x:
        return int((len(x)-len(x.replace(y,'')))/len(y))
    return int(0)
def getEnclosed(x,ls):
    if type(x) is not list:
        x = [x]
    z = ''
    lsN = []
    countLs = [int(0),int(0)]
    for i in range(0,len(x)):
        line = x[i]
        zN = ''
        for j in range(0,len(line)):
            li = line[j]
            zN = zN+li
            if li in ls:
                if countLs[0] == 0:
                    z = ls[0]
                    
                    lsN = []
                countLs[findIt(li,ls)] += 1
                if countLs[0] == countLs[1]:
                    lsN.append(zN)
                    z = z + zN
                    return lsN,z
        z = z + zN
        lsN.append(line)
    
def findIt(x,ls):
    for i in range(0,len(ls)):
        if x == ls[i]:
            return i
    return False
def getBrackDiff():
    return lsBrack[0]-lsBrack[1]
def trackTabs(x):
    ls = ['{','}']
    found = findIt(x,['{','}'])
    if found != False:
        lsBrack[found] += 1
        lsBrack[2] = getAsMany('\t',getBrackDiff())
def getAsMany(x,k):
    z = ''
    for i in range(0,k):
        z = adSt(z,x)
    return z
def getCurr(ls):
    if len(ls) >0:
        return ls[-1]
    return None
def removeFirst(x,y):
    if y in x:
        return x.replace(x.split(y)[0]+y,''),x.split(y)[0]
    return '',x
def checkSyntax(x):
    names = syntax['names']
    for i in range(0,len(names)):
            name = names[i]
            for j in range(0,len(syntax[name])):
                now = syntax[name][j]
                if x == now:
                    return name
    return False
def checkNames(x):
    ls = ['interface','abstract','library','contract']
    for i in range(0,len(ls)):
        if 'names' not in lineTrack:
            lineTrack[ls[i]] = {'names':[]}
        if x in lineTrack[ls[i]]['names']:
            return ls[i]
    return False
def checkDeclared(x):
    if x in lineTrack['declared']:
        return True
    return False
def linesToString(ls):
    all = ''
    for i in range(0,len(ls)):
        all = all + ls[i]
    return all
def checkCurr(n):
    trackTabs(n)
    add = lsBrack[2]
    if n == ' ':
        #lineTrack['currLineChop'],chop = #removeFirst(lineTrack['currLineChop'],' ')
        chop = eatAll(lineTrack['newLine'][1:],[' ','','\t','\n'])
        lineTrack['currPhrase'] = chop+n
        currLineLs.append(lineTrack['currLineChop'])
        lineTrack['currLineAttrib'].append(False)
        if lsCheck(chop,heads) != False:
            lineTrack['currLineAttrib'][-1] = lsCheck(chop,heads)
            lineTrack[lsCheck(chop,heads)]['names'][-1] = chop
        if lsCheck(chop,heads) != False:
            lineTrack['currLineAttrib'][-1] = lsCheck(chop,heads)
            lineTrack[lsCheck(chop,heads)]['names'][-1] = chop
        if lineTrack['currLineAttrib'][-1] == False:
            lineTrack['currLineAttrib'][-1] = lsCheck(chop,subHeads)
            if lsCheck(chop,subHeads) !=False:
                lineTrack[lsCheck(chop,subHeads)]['names'][-1] = chop
                
        if lineTrack['currLineAttrib'][-1] == False:
            lineTrack['currLineAttrib'][-1] = checkSyntax(chop)
        if lineTrack['currLineAttrib'][-1] == False:
            lineTrack['currLineAttrib'][-1] = checkNames(chop)
        if lineTrack['currLineAttrib'][-1] == False:
            lineTrack['currLineAttrib'][-1] = checkDeclared(chop)
    elif n == '{':
        n = n+'\n'+add
    elif n == '}':
        n = add+n+'\n'
    elif n == ';':
        n = ';\n'+add
    
    if newLine == None:
       changeGlob('currLineChop',n)
    else:
        lineTrack['currLineChop'] = lineTrack['currLineChop'] + n
def getAtts(x,lsN):
    if ' is ' in x:
        atts = x.split(' is ')[1].split('{')[0].replace(' ,',',').split(',')
        for i in range(0,len(atts)):
            att = eatAll(atts[i],[' ','','\t','\n'])
            if att not in lsN:
                lsN.append(att)
    return lsN
def getHeadsName(x,y):
    na = x.split(y+' ')[1].split(' ')[0]
    if y not in lineTrack:
        lineTrack[y] = {'names':[]}
    if na not in lineTrack[y]['names']:
        lineTrack[y]['names'].append(na)
    if na not in lineTrack[y]:
        lineTrack[y][na] = {'attributes':[],'names':{},'funcs':[]}
    lineTrack[y][na]['attributes'] = getAtts(x,lineTrack[y][na]['attributes'])
    return na
def getTagName(x,y):
    na = eatAll(x.split(y)[1],[' ',';','','\n','\t'])
    if y not in lineTrack:
        lineTrack[y] = {'names':[]}
    lineTrack[y]['names'].append(na)
    return na
def getCurrJs():
    return lineTrack[getCurrSect()][getCurrName()]
def getCurrName():
    return lineTrack[getCurrSect()]['names'][-1]
def currNameCheck():
    if 'currSect' in lineTrack:
        if 'names' in lineTrack[getCurrSect()]:
            if len(lineTrack[getCurrSect()]['names'])>0:
                return True
    return False
def getCurrAtts():
    return lineTrack[getCurrSect()][getCurrName()]['attributes']
def inputCurrAtt(x):
    if x not in getCurrAtts():
        lineTrack[getCurrSect()][getCurrName()]['attributes'].append(x)
    else:
        input(x+' is already used as a attribute name for '+getCurrName())
def getCurrFuncs():
    return lineTrack[getCurrSect()][getCurrName()]['funcs']
def inputCurrFuncs(x):
    if x not in getCurrFuncs():
        lineTrack[getCurrSect()][getCurrName()]['funcs'].append(x)
    else:
        input(x+' is already used as a function name for '+getCurrName())
def getUsedLibrary(x):
    inputCurrAtt(x.split('using ')[1].split(' ')[0])
def getModName(x):
    na = x.split('modifier ')[1].split(' ')[0]
    if 'modifier' not in lineTrack[getCurrSect()][getCurrName()]:
        lineTrack[getCurrSect()][getCurrName()]['modifier'] = {'names':[],'funcs':[]}
    if na not in lineTrack[getCurrSect()][getCurrName()]['modifier']['names']:
        lineTrack[getCurrSect()][getCurrName()]['modifier']['names'].append(na)
        enclosed = getEnclosed(lines[defs['currLineI']:],['{','}'])
        if enclosed not in lineTrack[getCurrSect()][getCurrName()]['modifier']['funcs']:
            lineTrack[getCurrSect()][getCurrName()]['modifier']['funcs'].append(enclosed)
        else:
            input(enclosed+' is already designated as '+lineTrack[getCurrSect()][newNa]['modifier']['names'][findIt(na,+lineTrack[getCurrSect()][newNa]['modifier']['funcs'])])
    else:
        input(na+' is already used as: '+lineTrack[getCurrSect()][newNa]['modifier']['funcs'][findIt(na,lineTrack['modifier']['names'])])
def getFuncName(x):
    inputCurrFuncs(x.split('function ')[1].split('(')[0])
def getConstVals(x,lsN):
    constVars = x.split('constructor(')[1].split(')')[0].split(',')
    if constVars != '':
        for i in range(0,len(constVars)):
            js = {}
            breakDown = constVars[i].split(' ')
            for j in range(0,len(breakDown)):
                if j == 0:
                    for c in range(0,len(syntax["type"])):
                        ty = syntax["type"][c]
                        if breakDown[j][:len(ty)] == ty:
                            js['type'] = breakDown[j]
                elif j == (len(breakDown)-1):
                    js['name'] = breakDown[j]
                    print(js)
                elif j != 0 and j !=  (len(breakDown)-1):
                    for c in range(0,len(syntax["names"])):
                        tyNa = syntax["names"][c]
                        ty = syntax[tyNa]
                        for k in range(0,len(ty)):
                            if breakDown[j][:len(ty[k])] == ty[k]:
                                js[tyNa] = breakDown[j]
            lsN.append(js)
            
    return  lsN
def cleanLsMod(x,ls):
    lsN = []
    if type(x) is not list:
        x = [x]
    for i in range(0,len(x)):
        lsN.append(eatAll(x[i],ls).replace('\t',''))
    return lsN
def getCurrSect():
    return lineTrack['currSect']
def inputCurrSect(x):
    lineTrack['currSect'] = x
def getCurrSub():
    return lineTrack['currSub']
def inputCurrSub(x):
    lineTrack['currSub'] = x
def getSyntax(x):
    for i in range(0,len(syntax['names'])):
        na = syntax['names'][i]
        naLs = syntax[na]
        for j in range(0,len(naLs)):
            currNa = naLs[j]
            #print(x,currNa)
            if x[:len(currNa)] == currNa:
                return na
    return False
def cleanLs(lsOg):
    lsN = []
    ls = ['',' ']
    for i in range(0,len(lsOg)):
        if lsOg[i] not in ls:
            lsN.append(lsOg[i])
    return lsN
def headCheck(x):
    heads = defs['heads']
    for i in range(0,len(heads)):
        head = heads[i]
        if x in lineTrack[head]['names']:
            #print(x,head)
            return head
    return False
def namesCheck(x):
    heads = defs['heads']
    for i in range(0,len(heads)):
        head = heads[i]
        na = lineTrack[head]['names']
        for j in range(0,len(na)):
            if x in lineTrack[head][na[j]]['names']:
                return head,na[j]
    return False
def declaredCheck(x):
    for i in range(0,len(defs['variables'])):
        if 'name' in defs['variables'][i]:
            if x == defs['variables'][i]['name']:
                return defs['variables'][i]
    return False
def addressCheck(x):
    if '0x' in x:
        if len(x) == len('0x05BFC3F8E9124eCBc762f841eC540DB3987AFE82'):
            return True
    else:
        if len(x) == len('05BFC3F8E9124eCBc762f841eC540DB3987AFE82'):
            return True
    return False

def getVariableInfo(og):

    parts = cleanLs(og.split(' '))
    js = {}
    lsN = []

    for i in range(0,len(parts)):
        partSyn = getSyntax(parts[i])
        if partSyn != False:
            js[partSyn] = parts[i]
            lsN.append(partSyn)
        elif i == len(parts)-1:
            js['name'] = parts[i]
            lsN.append('name')
        elif headCheck(parts[i]) != False:
            head = headCheck(parts[i])
            if i == 0:
                na = parts[-1].split(';')[0]
                if parts[i] not in lineTrack[head][parts[i]]['names']:
                    lineTrack[head][parts[i]]['names'][na] = {'address':""}
                    
            else:
                na = js['name']
                if parts[i] not in lineTrack[head][parts[i]]['names']:
                    lineTrack[head][parts[i]]['names'][na] = {'address':""}
            lsN.append(head)
            js[head] = parts[i]
        elif namesCheck(parts[i]) != False:
            head,na = namesCheck(parts[i])
            if '=' in og:
                if na in og.split('=')[1]:
                    lineTrack[head][na]['names'][parts[i]]['address'] = og.split('=')[1].split(na+'(')[1].split(')')[0]
            lsN.append(head)
            js[head] = parts[i]
        elif addressCheck(parts[i]) != False:
            if 'variabes' not in lineTrack[getCurrSect()][getCurrName()]:
                lineTrack[getCurrSect()][getCurrName()]['variables'] = {parts[i]:{'address':add}}
                                 
        else:
            if 'False' not in js:
                js['False'] = []
            js['False'].append(parts[i])
            lsN.append(None)
    return js,lsN
def getForWhileInfo(x):
    na = eatAll(x.split('(')[0],[' ',';','','\n','\t'])
    lenNa = len(na+'(')
    nothing,z = getEnclosed(lines[defs['currLineI']:],['(',')'])
    if 'loops' not in lineTrack[getCurrSect()][getCurrName()]:
        lineTrack[getCurrSect()][getCurrName()]['loops'] = {'for':{'vars':[],'funcs':[]},'while':{'vars':[],'funcs':[]}}
    lineTrack[getCurrSect()][getCurrName()]['loops'][na]['vars'].append(cleanLsMod(nothing[0].split(na+'(')[1].split(')')[0].split(';'),[' ',';','','\n','\t']))
    lsN,z = getEnclosed(lines[defs['currLineI']:],['{','}'])
    lineTrack[getCurrSect()][getCurrName()]['loops'][na]['funcs'].append(cleanLsMod(lsN[1:-1],[' ',';','','\n','\t']))
    defs['currLineI'] = defs['currLineI']+len(lsN)
def getStructName(x):
    na = eatAll(x.split('struct')[1].split('{')[0],['',' ','\n','\t'])
    if 'structs' not in lineTrack[getCurrSect()][getCurrName()]:
        lineTrack[getCurrSect()][getCurrName()]['structs'] = {'names':[]}
    if na not in  lineTrack[getCurrSect()][getCurrName()]['structs']['names']:
        lineTrack[getCurrSect()][getCurrName()]['structs']['names'].append(na)
        lineTrack[getCurrSect()][getCurrName()]['structs'][na] = {'variables':[]}
    lsN,z = getEnclosed(lines[defs['currLineI']:],['{','}'])
    varis = z.split(';')[1:-1]
    for i in range(0,len(varis)):
        va = eatAll(varis[i],['',' ','\n','\t']).split(' ')
        if type(va) is not list:
            va = [va]
        lineTrack[getCurrSect()][getCurrName()]['structs'][na]['variables'].append({'type':eatAll(va[0],['',' ','\n','\t']).replace('\t',''),'name':eatAll(va[-1],['',' ','\n','\t'])})
    defs['currLineI'] = defs['currLineI'] + len(lsN)
def getMappingName(x):
    na = eatAll(x.split('mapping')[1].split(';')[0].split(' ')[1],['',' ','\n','\t'])
    if 'mapping' not in lineTrack[getCurrSect()][getCurrName()]:
        lineTrack[getCurrSect()][getCurrName()]['mapping'] = {'names':[],"funcs":[]}
    if na not in  lineTrack[getCurrSect()][getCurrName()]['mapping']['names']:
        lineTrack[getCurrSect()][getCurrName()]['mapping']['names'].append(na)
        lineTrack[getCurrSect()][getCurrName()]['mapping'][na] = {'variables':[]}
    lsN,z = getEnclosed(lines[defs['currLineI']:],['(',')'])
    varis = z.split(';')
    for i in range(0,len(varis)):
        va = varis[i].split(' ')
        if type(va) is not list:
            va = [va]
        lineTrack[getCurrSect()][getCurrName()]['mapping'][na]['variables'].append({'type':eatAll(va[0],['',' ','\n','\t']),'name':eatAll(va[-1],['',' ','\n','\t'])})
    defs['currLineI'] = defs['currLineI'] + len(lsN)
def isDeclare(x):
    for i in range(0,len(syntax['type'])):
        syn = syntax['type'][i]
        if x[:len(syn)] == syn:
            getVaraibleInfo(x)
def isCurr(x):
    ls = ['tags','heads','subHeads','loops']
    for i in range(0,len(ls)):
        currLs = ls[i]
        for j in range(0,len(defs[currLs])):
            curr = defs[currLs][j]
            if x[:len(curr)] == curr:
                if i == 0:
                    na = getTagName(x,curr)
                elif i == 1:
                    inputCurrSect(curr)
                    getHeadsName(x,curr)
                elif i == 2:
                    inputCurrSub(curr)
                    if j == 0:
                        if 'constructor' not in getCurrJs():
                            lineTrack[getCurrSect()][getCurrName()]['constructor'] = {'vars':[],'attributes':[]}
                        lineTrack[getCurrSect()][getCurrName()]['constructor']['vars'] = getConstVals(x,lineTrack[getCurrSect()][getCurrName()]['constructor']['vars'])  
                        lineTrack[getCurrSect()][getCurrName()]['constructor']['attributes'] = getAtts(x,lineTrack[getCurrSect()][getCurrName()]['constructor']['attributes'])
                    elif j == 1:
                        if 'modifier' not in getCurrJs():
                            lineTrack[getCurrSect()][getCurrName()]['modifier'] = {'names':[],'funcs':[]}
                        getModName(x)
                    elif j == 2:
                        getUsedLibrary(x)
                    elif j == 3:
                        getFuncName(x)
                    elif j == 4:
                        getStructName(x)
                    elif j == 5:
                        getMappingName(x)
                elif i == 3:
                    getForWhileInfo(x)
                return True
    return False
def lsCheck(x,ls):
    if x in ls:
        return x
    return False
def ifComm(x):
    if x[:len('//')] == '//':
        if 'SPDX-License-Identifier' not in x:
            return False
        return x+'\n'
    return x
global lineTrack,currLineChop       
lsChangeGlob(['lines','lsBrack'],[reader('NeFiFeeManager.sol').split('\n'),[0,0,'']])
lineTrack = {}

changeGlob("syntax",{ "names":["type","modifiers","visibility","precedence","globalVariables","precedence"],
                     "modifiers":["pure","view","payable","constant","immutable","anonymous","indexed","virtual","override"],
                     "precedence":["assert","block","coinbase","difficulty","number","block;number","timestamp","block;timestamp","msg","data","gas","sender","value","gas price","origin","revert","require","keccak256","ripemd160","sha256","ecrecover","addmod","mulmod","cryptography","this","super","selfdestruct","balance","codehash","send"],
                     "visibility":["public","private","external","internal"],
                     "modifiers":["pure","view","payable","constant","anonymous","indexed"],
                     "globalVariables":['abi.decode(bytes memory encodedData, (...)) returns (...)', 'abi.encode(...) returns (bytes memory)', 'abi.encodePacked(...) returns (bytes memory)',  'abi.encodeWithSelector(bytes4 selector, ...) returns (bytes memory)', 'abi.encodeCall(function functionPointer, (...)) returns (bytes memory)','abi.encodeWithSelector(functionPointer.selector, (...))', 'abi.encodeWithSignature(string memory signature, ...) returns (bytes memory)', 'to abi.encodeWithSelector(bytes4(keccak256(bytes(signature)), ...)', 'bytes.concat(...) returns (bytes memory)', 'arguments to one byte array<bytes-concat>`', 'string.concat(...) returns (string memory)', 'arguments to one string array<string-concat>`', 'block.basefee (uint)', 'block.chainid (uint)', 'block.coinbase (address payable)', 'block.difficulty (uint)', 'block.gaslimit (uint)', 'block.number (uint)', 'block.timestamp (uint)', 'gasleft() returns (uint256)', 'msg.data (bytes)', 'msg.sender (address)', 'msg.sig (bytes4)', 'msg.value (uint)', 'tx.gasprice (uint)', 'tx.origin (address)', 'assert(bool condition)', 'require(bool condition)', 'for malformed input or error in external component)', 'require(bool condition, string memory message)', 'condition is false (use for malformed input or error in external component). Also provide error message.', 'revert()', 'revert(string memory message)', 'blockhash(uint blockNumber) returns (bytes32)', 'keccak256(bytes memory) returns (bytes32)', 'sha256(bytes memory) returns (bytes32)', 'ripemd160(bytes memory) returns (bytes20)', 'ecrecover(bytes32 hash, uint8 v, bytes32 r, bytes32 s) returns (address)', 'the public key from elliptic curve signature, return zero on error', 'addmod(uint x, uint y, uint k) returns (uint)', 'arbitrary precision and does not wrap around at 2**256. Assert that k != 0 starting from version 0.5.0.', 'mulmod(uint x, uint y, uint k) returns (uint)', 'with arbitrary precision and does not wrap around at 2**256. Assert that k != 0 starting from version 0.5.0.', "this (current contract's type)", 'super', 'selfdestruct(address payable recipient)', '<address>.balance (uint256)', '<address>.code (bytes memory)', '<address>.codehash (bytes32)', '<address payable>.send(uint256 amount) returns (bool)', 'returns false on failure', '<address payable>.transfer(uint256 amount)', 'type(C).name (string)', 'type(C).creationCode (bytes memory)', 'type(C).runtimeCode (bytes memory)', 'type(I).interfaceId (bytes4)', 'type(T).min (T)', 'type(T).max (T)'],"sectionHeader":['contract','library','interface','abstract contract'],"allVars":['contract','library','pragma solidity','import','interface','abstract contract','constructor','function','modify','SPDX-License-Identifier'],
                     "type":["address","uint","string","bytes",'bool']})
changeGlob('defs',{})
defs['tags'] = ['//:','pragma solidity']
defs['heads']=['contract','abstract contract','interface','library']
defs['subHeads']=['constructor','modifier','using','function','struct','mapping']
defs['loops'] = ['for(','while(']
defs['currLineI'] = 0
defs['variables'] = []
changeGlob('subHead',[])
changeGlob('currHead',[])
changeGlob('currLicense',[])
changeGlob('currPragma',[])
changeGlob('currLineLs',[])
changeGlob('currLine',None)
changeGlob('newLine',None)
changeGlob('lineTrack',{'declared':[],'variables':[],'currLineChop':"",'currLineAttrib':[],'currPrase':"",'newLine':""})
all = []
currLine = ''
currPhrase = ''
for i in range(0,len(lines)):
    while ifComm(lines[i]) == False:
        i += 1
    line = ifComm(lines[i])
    defs['currLineI'] = i
    changeGlob('currLine',eatAll(line,[' ','\n','\t','']))
    changeGlob('currLineLs',[])
    lineTrack['currPhrase'] = ''
    lineTrack['newLine'] = ''
    if isCurr(currLine) == False and currLine != None:
        if currNameCheck() == True:
            if getCurrName() == 'NeFiFeeManager' and currLine not in ['[',']','{','}']:
                if getVariableInfo(currLine) != False:
                    js,jsN = getVariableInfo(currLine)
                    defs['variables'].append(js)
    print(i)  
    i = defs['currLineI']
print('sdffdsfsd')
pen(lineTrack,'lineTrack.txt')
pen(str(defs['variables']).replace('},','},\n'),'all.txt')



from .abstract_rpcs import RPCBridge
from web3 import Web3
class w3Manager:
    def __init__(self):
        self.w3_management = {}
        self.default_rpc_js = get_default_rpc(Network_Name="Ethereum")  # Assuming get_default_rpc is defined elsewhere
        self.default_rpc_mgr = RPCBridge(rpc_js=self.default_rpc_js)  # Assuming get_new_rpc_bridge is defined elsewhere
        self.default_w3 = self.default_rpc_mgr.w3

    def add_to_list(self, address, w3=None, rpc=None,rpc_js=None):
        normalized_address = get_normalized_address(address)
        if not w3 and rpc:
            w3 = self.create_w3(address, rpc=rpc)
        self.w3_management[normalized_address] = {'w3': w3 or self.default_w3, 'rpc': rpc or self.default_rpc_mgr.rpc,'rpc_js':rpc_js or self.default_rpc_js}

    def retrieve_w3(self, address):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        return self.w3_management[normalized_address]['w3']
    def retrieve_rpc(self, address):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        return self.w3_management[normalized_address]['rpc']
    def retrieve_rpc_js(self, address):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        return self.w3_management[normalized_address]['rpc_js']
    def create_w3(self, address, rpc=None):
        if isinstance(rpc, dict):
            self.ammend_to_list(get_normalized_address(address), ['rpc_js'], [rpc])
            rpc = rpc.get('user_settings', rpc.get('rpc', self.default_rpc_mgr.rpc))
            if rpc and 'rpc' in rpc and isinstance(rpc,dict):
                rpc=rpc.get('rpc')
            
        elif rpc is None:
            rpc = self.default_rpc_mgr.rpc

        try:
            new_w3 = Web3(Web3.HTTPProvider(rpc))
            self.ammend_to_list(address, ['w3', 'rpc'], [new_w3, rpc])
            return new_w3
        except Exception as e:
            print(f"Error creating w3 from rpc {rpc}: {e}")
            self.add_to_list(address)  # Fallback to default w3
            return self.default_w3

    def ammend_to_list(self, address, keys, values):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        for i, key in enumerate(keys):
            self.w3_management[normalized_address][key] = values[i]



from abstract_utilities.path_utils import makeAllDirs
from abstract_utilities.type_utils import is_number, convert_to_number,make_list
from abstract_utilities.json_utils import closest_dictionary, find_matching_dicts,safe_read_from_json
from abstract_utilities.compare_utils import get_closest_match_from_list

from abstract_gui import AbstractWindowManager,sg,make_component,create_row_of_buttons,text_to_key,get_event_key_js
from web3 import Web3
import os
import json

def get_common_chains():
        return ['Arbitrum One', 'Avalanche C-Chain', 'Endurance Smart Chain Mainnet', 'Celo Mainnet', 'Cronos Mainnet', 'Elastos Smart Chain', 'Ethereum Mainnet', 'Fuse Mainnet', 'Gnosis', 'Huobi ECO Chain Mainnet', 'Hoo Smart Chain', 'IoTeX Network Mainnet', 'Catecoin Chain Mainnet', 'Polygon Mainnet', 'Moonriver', 'Nahmii Mainnet', 'OKXChain Mainnet', 'Harmony Mainnet Shard 0', 'PandoProject Mainnet', 'Smart Bitcoin Cash', 'Neon EVM Mainnet', 'Telos EVM Mainnet', 'Ubiq']

def strip_web(url:str):
    if isinstance(url,dict):
        url = url.get('url',url)        
    if url:
        if url.startswith("http://"):
            url = url.replace("http://", '', 1)
        elif url.startswith("https://"):
            url = url.replace("https://", '', 1)
        url = url.split('/')[0]
        return url

def get_total_values_list(rpc_list):
        new_list = []
        for rpc in rpc_list:
                for value in rpc.values():
                        if isinstance(value, (str, int)):
                                new_list.append(value)
        return new_list
def get_default_rpc_list():
    
    rpc_list =read_json(makeAllDirs(os.path.join(os.path.dirname(os.path.abspath(__file__)),'data','rpc_list.json')))
    return rpc_list['chains']


def separate_testnet_rpcs(rpc_list):
    """
    Separates testnet and non-testnet RPCs.
    """
    testnet_rpcs = [rpc for rpc in rpc_list if 'testnet' in rpc['name'].lower()]
    non_testnet_rpcs = [rpc for rpc in rpc_list if 'testnet' not in rpc['name'].lower()]
    return testnet_rpcs, non_testnet_rpcs

def get_closest_values(rpc_js,total_values_list):
    rpc_js_new = []
    for i,rpc_value in enumerate(rpc_js):

        value = get_closest_match_from_list(rpc_value, list(total_values_list),sorted_out=False,highest_out=True)
        rpc_js_new.append(value)

    return rpc_js_new

def get_testnet(testnet, rpc_list):
    return [rpc for rpc in rpc_list if ('testnet' in rpc['name'].lower().split(' ')) == testnet]

def get_rpc_urls(rpc_js):
    urls=[]
    rpc=''
    rpcs=[]
    if 'rpc' in rpc_js:
        rpcs = find_matching_dicts(dict_objs=rpc_js['rpc'],keys=['tracking'],values=['none']) or rpc_js['rpc']
    for rpc in rpcs:
        if 'url' in rpc:
            urls.append(rpc['url'])
    return urls

def get_explorers(rpc_js):
    urls=[]
    explorer=''
    explorers=[]
    if 'explorers' in rpc_js:
        explorers=rpc_js['explorers']
    for explorer in explorers:
        if 'url' in explorer:
            urls.append(explorer['url'])
    return urls

def get_default_rpc(Network_Name:str="Ethereum", rpc_list:list=None):
    if rpc_list == None:
        rpc_list =read_json(os.path.join(os.path.dirname(os.path.abspath(__file__)),'data','rpc_list.json'))
        rpc_list= rpc_list['chains']
    return closest_dictionary(dict_objs=rpc_list,values=[Network_Name])
def read_json(file_path):
    try:
        with open(file_path, 'r') as file:
            return json.load(file)
    except FileNotFoundError:
        return None
    except json.JSONDecodeError:
        raise ValueError(f"Invalid JSON in {file_path}")

def write_json(file_path, data):
    with open(file_path, 'w') as file:
        json.dump(data, file, indent=4)




from time import time
import setuptools
from setuptools import setup, find_packages

with open('README.md', "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='abstract_blockchain',
    version='0.0.2.03',
    author='putkoff',
    author_email='partners@abstractendeavors.com',
    description='The Abstract Blockchain package provides a collection of modules designed to simplify interaction with blockchain networks, smart contracts, and related components. It offers tools for managing RPC parameters, working with smart contract ABIs, and facilitating user-friendly interactions through graphical user interfaces (GUIs).',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    url='https://github.com/AbstractEndeavors/abstract_blockchain',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development',
        'Topic :: Utilities',
        'Intended Audience :: Financial and Insurance Industry'
    ],
    keywords='database, automation, API, testing, cryptography',  # This is how you add keywords
    install_requires=[
        'PySimpleGUI>=4.60.5', 'abstract_security>=0.0.1', 'abstract_webtools>=0.1.5.4', 'abstract_gui>=0.0.62.03', 'web3>=6.9.0', 'abstract_utilities>=0.2.2.34', 'requests>=2.31.0', 'hexbytes>=0.3.1'
    ],
    package_dir={"": "src"},
    package_data={'data': ['rpc_list.json']},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    setup_requires=['wheel'],
)



from web3 import Web3
import sys
import os
import json
import PySimpleGUI as sg
import codecs
import requests
from hexbytes import HexBytes
from web3 import Web3, HTTPProvider
def avgGasPriceEstd():
    pending_transactions = web3.provider.make_request("parity_pendingTransactions", [])
    gas_prices,gases = [],[]
    for tx in pending_transactions["result"[:10]]:
        gas_prices.append(int((tx["gasPrice"]),16))
        gases.append(int((tx["gas"])))
    return statistics.median(gas_prices)
def avgGasPriceEst():
    req = requests.get('https://ethgasstation.info/json/ethgasAPI.json')
    t = json.loads(req.content)
    gas_price1 = web3.eth.gasPrice
    return int(t[gasSpeed]*(10**8))
def specificverifyInput(inType,inVar,outType,outVar,fun,ask):
    asky = ''
    layout = [[sg.Text(ask), sg.Yes()],
            [sg.Text('choose anther '+str(outVar)+' input:'), sg.No()],
            [sg.Yes(), sg.No()]]
    window = sg.Window('Window Title', layout)
    while asky == '':             
        event, values = window.read()
        print(values)
        if event in (sg.WIN_CLOSED, 'Cancel'):
            asky = False
        window.close()
        asky = True
    window.close()  
    return asky
def specificAskInput(inType,inVar,fun,ask):
    asky = ''
    y = asky
    layout = [
        [sg.Text(inType + ' '+ inVar), sg.Push(), sg.Input('', key=inType)],
        [sg.Text('Input Verification'), sg.Push(), sg.Input('', disabled=True, key='verify')],
        [sg.OK('OK'),sg.Button('Show'), sg.Button('Exit'),]]
    window = sg.Window(fun, layout)
    while asky == '':
        event, values = window.read()
        if event == sg.WIN_CLOSED or event == 'Exit':
            asky = 'exit'
        if event == sg.OK() or event == 'OK':
            asky = values[inType]
        elif event == 'Show':
            x_string = values[inType]
            if 'address' in inType:
                y = addressVerify(inType,x_string)
                if y == False:
                    y = 'checkSum was unable to verify '+x_string+' '+inType
                else:
                    y = inType+' is good to go!'
            if 'uint' in inType:
                y = uintVerify(inVar,x_string)
                if y == False:
                    y = x_string+'is not a valid '+inType+' input'
                else:
                    y = inType+' is good to go!'
            window['verify'].update(value=y)
    window.close()
    return asky
def askList(nets):
    y = []
    sg.set_options(suppress_raise_key_errors=False, suppress_error_popups=False, suppress_key_guessing=False)
    layout = [[sg.Text('functions:'),sg.Combo(nets, key='functions',default_value=nets[0]), sg.Push()],
    [sg.OK('OK'),sg.Button('Show'),sg.Button('Exit')]]
    window = sg.Window('Window Title', layout, finalize=True)
    while True:             # Event Loop
        event, values = window.read()
        print(event, values)
        if event == sg.OK() or event == 'OK':
            return values['functions']
        if event == sg.WIN_CLOSED or event == 'Exit':
            break
    window.close()
def specificTextInput(inType,inVar,qu,outType,outVar,fun,ask):
    layout = [
        [sg.Text(inType + ' '+ inVar), sg.Push(), sg.Input('', key='x')],
        [sg.Text(outType+' '+outVar), sg.Push(), sg.Input('', disabled=True, key='x+2')],
        [sg.Button('Show'), sg.Button('Exit')]]
    window = sg.Window(fun, layout)
    while True:  # Event Loop
        event, values = window.read()
        if event == sg.WIN_CLOSED or event == 'Exit':
            break
        elif event == 'Show':
            x_string = values['x']
            try:
               y = ans = grab.tryCheckSum(x_string)

            except ValueError:
                y = "Wrong number !!!"
            window['x+2'].update(value=y)
    window.close()
def simpleBool(ask):
    asky = ''
    layout = [[sg.Text(ask)],
            [sg.Yes(), sg.No()]]
    window = sg.Window('Window Title', layout)
    while asky == '':             
        event, values = window.read()
        if event in (sg.WIN_CLOSED, 'Cancel'):
            asky = False
        window.close()
        asky = True
    window.close()  
    return asky
def mkBool(x):
    if str(x).lower() in ['true','t','yes','y','0']:
        return str(0)
    return str(1)    
def getType(ty,x):
    if 'int' in ty:
        return int(x)
    elif 'bytes' in ty:
        return kek(x)
    elif 'address' in ty:
        return w3.toChecksumAddress(str(x))
    elif 'bool' in ty:
        return bool(str(x))
    else:
        return str(x)
def cleanLs(ls):
    lsN = []
    for k in range(0,len(ls)):
        if ls[k] != '':
            lsN.append(ls[k])
    return lsN    
def ifLs(ty,x):
    x = cleanLs(mkLs(x.split(',')))
    for k in range(0,len(x)):
        x[k] = getType(ty,x[k])
    return x
def addressVerify(va,y):
    return tryCheckSum(str(y))
def uintVerify(va,y):
    return f.isInt(y)
def checkSum(x):
        return w3.toChecksumAddress(x)
def tryCheckSum(x):
    try:
        y = checkSum(x)
        return y
    except:
        return False
def safeSplit(x,y,k):
    if y in x:
        z = x.split(y)
        if len(z) > k:
            return z
def isAddress(x):
    if x[:len('address')] == 'address':
        return True
def isBool(x):
    if x[:len('bool')] == 'bool':
        return True
def isString(x):
    if x[:len('string')] == 'string':
        return True
def isBytes(x):
    if x[:len('bytes')] == 'bytes':
        return True
def isLs(ls):
    if type(ls) is list:
        return True
    return False
def isList(x):
    if x[-1] == ']':
        if x[:-1].split('[')[-1] == '':
            return [x],'*',0
        lsN = []
        for i in range(0,int(x[:-1].split('[')[-1])):
            lsN.append(x)    
        return lsN,i,0
    return x,1,0
def isUintSt(x):
    if x[:len('uint')] == 'uint' and str(x).lower() in ['uint8','uint16','uint32','uint64','uint128','uint256']:
        return True
    return False
def isIntSt(x):
    if x[:len('int')] == 'int' and str(x).lower() in ['int8','int16','int32','int64','int128','int256']:
        return True
    return False
def isInt(x):
  if type(x) is int:
    return True
  return False
def lsNum():
    return str('1,2,3,4,5,6,7,8,9,0').split(',')
def isNum(x):
    if isInt(x):
        return True
    for k in range(0,len(str(x))):
        if str(x)[k] not in lsNum():
            return False
    return True
def indetLoop(ty,k):
    tot = input('indeterminate '+str(ty)+' list, how many inputs will you be appending to it? (will inquire again at this input interval, otherise, inquiry will repeat every loop)')
    deff = f.isInt(tot)
    if deff == False:
        tot = k + 1
    return tot,deff
def listLoop(k,tot):
    print(str(k)+' out of '+str(tot)+' inputs in this list')
def denyInp(ty,va,fun,ans,exp):
    print('your '+str(ty)+' input for '+str(va)+' in function '+str(fun)+'; '+str(ans)+' was denied for '+exp)
def verify(ty,va,fun,ans):
    return specificverifyInput(ty,va,ty,ans,fun,'your '+str(ty)+' input for '+str(va)+' in functin '+str(fun)+' is '+str(ans)+' that ok?')
def askInput(ty,va,fun):
    return specificAskInput(ty,va,fun,'please input '+str(ty)+' for variable '+str(va)+' in function'+str(fun)+':')
def askBool(ty,va,fun):
    return specificAskInput(ty,va,fun,'please input '+str(ty)+' for variable '+str(va)+' in function'+str(fun)+':')
def inputUint(ty,va,fun):
    while isUintSt(ty):
        ast = False
        print('input will be multiplied by 10^'+str(int(18))+' if * is added to the end of the input)')
        ans = askInput(ty,va,fun)
        if ans[-1] == '*':
            ast,ans = True,ans[:-1]
        if isNum(ans):
            if ast == True:
                ans = int(float(int(ans))*float(str('1e'+str(int(18)))))
            if verify(ty,va,fun,ans):
                ifVarisLsApp(str(ans))
                return
def inputInt(ty,va,fun):
    while isIntSt(ty):
        ast = False
        print('input will be multiplied by 10^'+str(int(8))+' if * is added to the end of the input)')
        ans = askInput(ty,va,fun)
        if ans[-1] == '*':
            ast,ans = True,ans[:-1]
        if isNum(ans):
            if ast == True:
                ans = int(float(int(ans))*float(str('1e'+str(int(8)))))
            if verify(ty,va,fun,ans):
                ifVarisLsApp(str(ans))
                return
def ifVarisLsApp(x):
    if type(varis['inputs']['inpCurr'][-1]) is list:
        varis['inputs']['inpCurr'][-1].append(x)
    else:
        varis['inputs']['inpCurr'][-1] = x
def inputAddress(ty,va,fun):
    while isAddress(ty):
        ans = tryCheckSum(str(askInput(ty,va,fun)))
        if ans != False:
            if verify(ty,va,fun,ans) == True:
                ifVarisLsApp(ans)
                return 
        else:
            denyInp(ty,va,fun,ans,' bad checksum address')
def inputBool(ty,va,fun):
    while isBool(ty):
        ans = askBool(ty,va,fun)
        if verify(ty,va,fun,ans):
            if ans == True:
                ifVarisLsApp(ans)
            else:
                ifVarisLsApp(ans)
            return
def inputString(ty,va,fun):
    while isString(ty):
        ans = str(askInput(ty,va,fun))
        if verify(ty,va,fun,ans) == True:
            ifVarisLsApp(ans)
            return
def inputBytes(ty,va,fun):
    while isBytes(ty):
        ans = str(askInput(ty,va,fun))
        if verify(ty,va,fun,ans) == True:
            ifVarisLsApp(ans)
            return
def askAll(ty,va,fun):
    varis['inputs']['typeCurr'].append([ty,va,fun])
    varis['inputs']['inpCurr'].append(ty)
    deff = True
    tyAc,tot,cou = isList(ty)
    if type(ty) is list:
        if tot == '*':
            tot,deff = indetLoop(ty,cou)
        else:
            print('input requres list of a length '+str(tot)+'; the input request will repeat '+str(tot)+' times')
    while cou < tot:
        inputAddress(ty,va,fun)
        inputBool(ty,va,fun)
        inputString(ty,va,fun)
        inputBytes(ty,va,fun)
        inputUint(ty,va,fun)
        inputInt(ty,va,fun)
        cou += 1
        listLoop(cou,tot)
        if deff == False:
            tot = indetLoop(ty,cou)
def ifInput(js):
    lsType = ['address','uint','bool','bytes','string']
    if 'inputs' in js:

        inps = js['inputs']
        for i in range(0,len(inps)):
            if ' ' in inps[i]:
                askAll(inps[i].split(' ')[0],inps[i].split(' ')[-1],js['name'])
            else:
                 askAll(inps[i],inps[i],js['name'])
def homeIt():
	changeGlob("home",os.getcwd())
	slash = "//"
	if "//" not in str(home):
		slash = "/"
		changeGlob("slash",slash)
	return home,slash
def pen(paper, place):
	with open(place, "w") as f:
		f.write(str(paper))
		f.close()
		return
def isHex(x):
	try:
		z = x.hex()
		return True
	except:
		return False
def printHex(x):
	if isHex(x):
		return x.hex()
	return text
def changeGlob(x,v):
	globals()[x] = v
def readerC(file):
	with open(file,"r" ,encoding="utf-8-sig") as f:
		text = f.read()
		return text
def mkLs(ls):
  if type(ls) is not list:
    ls = [ls]
  return ls
def getCodex(x):
    b = get_hex_data(x)
    return codecs.decode(b, 'UTF-8')
def kek(x):
    st = '"'+str(x)+'"'
    return w3.keccak(text=str(x)).hex()
def read_hex(hb):
    h = "".join(["{:02X}".format(b) for b in hb])
    return h
def get_hex_data(x):
    n = len(x)
    hex = x
    return int(hex, n)



from .abstract_rpcs import RPCBridge
from .abstract_w3_manager import w3Manager
from abstract_security import get_env_value
from abstract_utilities import make_list,safe_json_loads,safe_dump_to_file,safe_read_from_json,create_and_read_json
from abstract_utilities.list_utils import remove_from_list
from abstract_utilities.path_utils import makeAllDirs
from .rpc_functions import *
import requests
import json
import os
from .abstract_rpcs import RPCBridge
from web3 import Web3
class w3Manager:
    def __init__(self):
        self.w3_management = {}
        self.default_rpc_js = get_default_rpc(Network_Name="Ethereum")  # Assuming get_default_rpc is defined elsewhere
        self.default_rpc_mgr = RPCBridge(rpc_js=self.default_rpc_js)  # Assuming get_new_rpc_bridge is defined elsewhere
        self.default_w3 = self.default_rpc_mgr.w3

    def add_to_list(self, address, w3=None, rpc=None,rpc_js=None):
        normalized_address = get_normalized_address(address)
        if not w3 and rpc:
            w3 = self.create_w3(address, rpc=rpc)
        self.w3_management[normalized_address] = {'w3': w3 or self.default_w3, 'rpc': rpc or self.default_rpc_mgr.rpc,'rpc_js':rpc_js or self.default_rpc_js}

    def retrieve_w3(self, address):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        return self.w3_management[normalized_address]['w3']
    def retrieve_rpc(self, address):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        return self.w3_management[normalized_address]['rpc']
    def retrieve_rpc_js(self, address):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        return self.w3_management[normalized_address]['rpc_js']
    def create_w3(self, address, rpc=None):
        if isinstance(rpc, dict):
            self.ammend_to_list(get_normalized_address(address), ['rpc_js'], [rpc])
            rpc = rpc.get('user_settings', rpc.get('rpc', self.default_rpc_mgr.rpc))
            if rpc and 'rpc' in rpc and isinstance(rpc,dict):
                rpc=rpc.get('rpc')
            
        elif rpc is None:
            rpc = self.default_rpc_mgr.rpc

        try:
            new_w3 = Web3(Web3.HTTPProvider(rpc))
            self.ammend_to_list(address, ['w3', 'rpc'], [new_w3, rpc])
            return new_w3
        except Exception as e:
            print(f"Error creating w3 from rpc {rpc}: {e}")
            self.add_to_list(address)  # Fallback to default w3
            return self.default_w3

    def ammend_to_list(self, address, keys, values):
        normalized_address = get_normalized_address(address)
        if normalized_address not in self.w3_management:
            self.add_to_list(address)
        for i, key in enumerate(keys):
            self.w3_management[normalized_address][key] = values[i]
w3_mgr = w3Manager()            
apiCallDesc = {'multiTopic': {'description': 'searchtopics', 'example': 'https://api.etherscan.io/api?module=logs&action=getLogs&fromBlock=12878196&toBlock=12879196&topic0=0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef&topic0_1_opr=and&topic1=0x0000000000000000000000000000000000000000000000000000000000000000&page=1&offset=1000&apikey=YourApiKeyToken', 'pieces': {'module': ['logs'], 'action': ['getLogs'], 'fromBlock': [], 'toBlock': [], 'topic0': [], 'topic0_1_opr': ['and'], 'topic1': [], 'page': [], 'offset': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['fromBlock', 'toBlock', 'topic0', 'topic1', 'page', 'offset'], 'title': 'multitopic'},
               'topicSearch': {'description': 'getbytopic', 'title': 'topicsearch', 'pieces': {'module': ['logs'], 'action': ['getLogs'], 'fromBlock': [], 'toBlock': ['latest'], 'address': [], 'topic0': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['fromBlock', 'address', 'topic0'], 'example': 'https://api.etherscan.io/api?module=logs&action=getLogs&fromBlock=379224&toBlock=latest&address=0x33990122638b9132ca29c723bdf037f1a891a70c&topic0=0xf63780e752c6a54a94fc52715dbc5518a3b4c3c2833d301a204226548a2a8545&apikey=YourApiKeyToken'},
               'EthBalForSnglAddrs': {'description': 'GetEtherBalanceforasingleAddress\nexample:\n/api?module=account&action=balance&address=0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae&tag=latest&apikey=YourApiKeyToken', 'title': 'GetEtherBalanceforasingleAddress', 'pieces': {'module': ['account'], 'action': ['balance'], 'address': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=account&action=balance&address=0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae&tag=latest&apikey=YourApiKeyToken'}, 'HistEthBalForSnglAddrsByBlkno': {'description': 'GetHistoricalEtherBalanceforasingleAddressByBlockNo\nexample:\n/api?module=account&action=balancehistory&address=0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae&blockno=8000000&apikey=YourApiKeyToken', 'title': 'GetHistoricalEtherBalanceforasingleAddressByBlockNo', 'pieces': {'module': ['account'], 'action': ['balancehistory'], 'address': [], 'blockno': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address', 'blockno'], 'example': '/api?module=account&action=balancehistory&address=0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae&blockno=8000000&apikey=YourApiKeyToken'}, 'EthBalForMulAddrsesInSnglCall': {'description': 'GetEtherBalanceformultipleAddressesinasinglecall\nexample:\n/api?module=account&action=balancemulti&address=0xddbd2b932c763ba5b1b7ae3b362eac3e8d40121a,0x63a9975ba31b0b9626b34300f7f627147df1f526,0x198ef1ec325a96cc354c7266a038be8b5c558f67&tag=latest&apikey=YourApiKeyToken', 'title': 'GetEtherBalanceformultipleAddressesinasinglecall', 'pieces': {'module': ['account'], 'action': ['balancemulti'], 'address': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=account&action=balancemulti&address=0xddbd2b932c763ba5b1b7ae3b362eac3e8d40121a,0x63a9975ba31b0b9626b34300f7f627147df1f526,0x198ef1ec325a96cc354c7266a038be8b5c558f67&tag=latest&apikey=YourApiKeyToken'}, 'LstOfNrmlTxnsByAddrs': {'description': 'GetalistofNormalTransactionsByAddress\nexample:\n/api?module=account&action=txlist&address=0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae&startblock=0&endblock=99999999&sort=asc&apikey=YourApiKeyToken', 'title': 'GetalistofNormalTransactionsByAddress', 'pieces': {'module': ['account'], 'action': ['txlist'], 'address': [], 'startblock': [], 'endblock': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address', 'startblock', 'endblock'], 'example': '/api?module=account&action=txlist&address=0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae&startblock=0&endblock=99999999&sort=asc&apikey=YourApiKeyToken'}, 'LstOfIntlTxnsByAddrs': {'description': 'GetalistofInternalTransactionsbyAddress\nexample:\n/api?module=account&action=txlistinternal&address=0x2c1ba59d6f58433fb1eaee7d20b26ed83bda51a3&startblock=0&endblock=2702578&sort=asc&apikey=YourApiKeyToken', 'title': 'GetalistofInternalTransactionsbyAddress', 'pieces': {'module': ['account'], 'action': ['txlistinternal'], 'address': [], 'startblock': [], 'endblock': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address', 'startblock', 'endblock'], 'example': '/api?module=account&action=txlistinternal&address=0x2c1ba59d6f58433fb1eaee7d20b26ed83bda51a3&startblock=0&endblock=2702578&sort=asc&apikey=YourApiKeyToken'}, 'IntlTxnsByTxnHash': {'description': 'GetInternalTransactionsbyTransactionHash\nexample:\n/api?module=account&action=txlistinternal&txhash=0x40eb908387324f2b575b4879cd9d7188f69c8fc9d87c901b9e2daaea4b442170&apikey=YourApiKeyToken', 'title': 'GetInternalTransactionsbyTransactionHash', 'pieces': {'module': ['account'], 'action': ['txlistinternal'], 'txhash': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['txhash'], 'example': '/api?module=account&action=txlistinternal&txhash=0x40eb908387324f2b575b4879cd9d7188f69c8fc9d87c901b9e2daaea4b442170&apikey=YourApiKeyToken'}, 'IntlTxnsByBlkRng': {'description': 'GetInternalTransactionsbyBlockRange\nexample:\n/api?module=account&action=txlistinternal&startblock=0&endblock=2702578&page=1&offset=10&sort=asc&apikey=YourApiKeyToken', 'title': 'GetInternalTransactionsbyBlockRange', 'pieces': {'module': ['account'], 'action': ['txlistinternal'], 'startblock': [], 'endblock': [], 'page': [], 'offset': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startblock', 'endblock', 'page', 'offset'], 'example': '/api?module=account&action=txlistinternal&startblock=0&endblock=2702578&page=1&offset=10&sort=asc&apikey=YourApiKeyToken'}, 'LstOfErc20-TknTransferEventsByAddrs': {'description': 'GetalistofERC20-TokenTransferEventsbyAddress\nexample:\n/api?module=account&action=tokentx&address=0x4e83362442b8d1bec281594cea3050c8eb01311c&startblock=0&endblock=999999999&sort=asc&apikey=YourApiKeyToken', 'title': 'GetalistofERC20-TokenTransferEventsbyAddress', 'pieces': {'module': ['account'], 'action': ['tokentx'], 'address': [], 'startblock': [], 'endblock': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address', 'startblock', 'endblock'], 'example': '/api?module=account&action=tokentx&address=0x4e83362442b8d1bec281594cea3050c8eb01311c&startblock=0&endblock=999999999&sort=asc&apikey=YourApiKeyToken'}, 'LstOfErc721-TknTransferEventsByAddrs': {'description': 'GetalistofERC721-TokenTransferEventsbyAddress\nexample:\n/api?module=account&action=tokennfttx&address=0x6975be450864c02b4613023c2152ee0743572325&startblock=0&endblock=999999999&sort=asc&apikey=YourApiKeyToken', 'title': 'GetalistofERC721-TokenTransferEventsbyAddress', 'pieces': {'module': ['account'], 'action': ['tokennfttx'], 'address': [], 'startblock': [], 'endblock': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address', 'startblock', 'endblock'], 'example': '/api?module=account&action=tokennfttx&address=0x6975be450864c02b4613023c2152ee0743572325&startblock=0&endblock=999999999&sort=asc&apikey=YourApiKeyToken'}, 'LstOfBlksMinedByAddrs': {'description': 'GetlistofBlocksMinedbyAddress\nexample:\n/api?module=account&action=getminedblocks&address=0x9dd134d14d1e65f84b706d6f205cd5b1cd03a46b&blocktype=blocks&apikey=YourApiKeyToken', 'title': 'GetlistofBlocksMinedbyAddress', 'pieces': {'module': ['account'], 'action': ['getminedblocks'], 'address': [], 'blocktype': ['blocks'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=account&action=getminedblocks&address=0x9dd134d14d1e65f84b706d6f205cd5b1cd03a46b&blocktype=blocks&apikey=YourApiKeyToken'}, 'GetContractABIforVerifiedContractSourceCodes': {'description': 'GetContractABIforVerifiedContractSourceCodes\nexample:\n/api?module=contract&action=getabi&address=0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413&apikey=YourApiKeyToken', 'title': 'GetContractABIforVerifiedContractSourceCodes', 'pieces': {'module': ['contract'], 'action': ['getabi'], 'address': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=contract&action=getabi&address=0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413&apikey=YourApiKeyToken'}, 'CntrctSrcCodeForVerifiedCntrctSrcCodes': {'description': 'GetContractSourceCodeforVerifiedContractSourceCodes\nexample:\n/api?module=contract&action=getsourcecode&address=0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413&apikey=YourApiKeyToken', 'title': 'GetContractSourceCodeforVerifiedContractSourceCodes', 'pieces': {'module': ['contract'], 'action': ['getsourcecode'], 'address': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=contract&action=getsourcecode&address=0xBB9bc244D798123fDe783fCc1C72d3Bb8C189413&apikey=YourApiKeyToken'}, 'HeckCntrctExecutionStatus': {'description': 'CheckContractExecutionStatus(iftherewasanerrorduringcontractexecution)\nexample:\n/api?module=transaction&action=getstatus&txhash=0x15f8e5ea1079d9a0bb04a4c58ae5fe7654b5b2b4463375ff7ffb490aa0032f3a&apikey=YourApiKeyToken', 'title': 'CheckContractExecutionStatus(iftherewasanerrorduringcontractexecution)', 'pieces': {'module': ['transaction'], 'action': ['getstatus'], 'txhash': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['txhash'], 'example': '/api?module=transaction&action=getstatus&txhash=0x15f8e5ea1079d9a0bb04a4c58ae5fe7654b5b2b4463375ff7ffb490aa0032f3a&apikey=YourApiKeyToken'}, 'HeckTxnReceiptStatus': {'description': 'CheckTransactionReceiptStatus(OnlyapplicableforPostByzantiumforktransactions)\nexample:\n/api?module=transaction&action=gettxreceiptstatus&txhash=0x513c1ba0bebf66436b5fed86ab668452b7805593c05073eb2d51d3a52f480a76&apikey=YourApiKeyToken', 'title': 'CheckTransactionReceiptStatus(OnlyapplicableforPostByzantiumforktransactions)', 'pieces': {'module': ['transaction'], 'action': ['gettxreceiptstatus'], 'txhash': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['txhash'], 'example': '/api?module=transaction&action=gettxreceiptstatus&txhash=0x513c1ba0bebf66436b5fed86ab668452b7805593c05073eb2d51d3a52f480a76&apikey=YourApiKeyToken'},
 'Blk&UncleRewByBlkno': {'description': 'GetBlockAndUncleRewardsbyBlockNo\nexample:\n/api?module=block&action=getblockreward&blockno=2165403&apikey=YourApiKeyToken', 'title': 'GetBlockAndUncleRewardsbyBlockNo', 'pieces': {'module': ['block'], 'action': ['getblockreward'], 'blockno': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['blockno'], 'example': '/api?module=block&action=getblockreward&blockno=2165403&apikey=YourApiKeyToken'}, 'EstBlkCountdownTimeByBlkno': {'description': 'GetEstimatedBlockCountdownTimebyBlockNo\nexample:\n/api?module=block&action=getblockcountdown&blockno=9100000&apikey=YourApiKeyToken', 'title': 'GetEstimatedBlockCountdownTimebyBlockNo', 'pieces': {'module': ['block'], 'action': ['getblockcountdown'], 'blockno': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['blockno'], 'example': '/api?module=block&action=getblockcountdown&blockno=9100000&apikey=YourApiKeyToken'}, 'BlkNumByTimestamp': {'description': 'GetBlockNumberbyTimestamp\nexample:\n/api?module=block&action=getblocknobytime&timestamp=1578638524&closest=before&apikey=YourApiKeyToken', 'title': 'GetBlockNumberbyTimestamp', 'pieces': {'module': ['block'], 'action': ['getblocknobytime'], 'timestamp': [], 'closest': ['before'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['timestamp'], 'example': '/api?module=block&action=getblocknobytime&timestamp=1578638524&closest=before&apikey=YourApiKeyToken'}, 'EventLogsFromBlkNum': {'description': 'GetEventLogsfromblocknumber379224tolatestBlock,wherelogaddress=0x33990122638b9132ca29c723bdf037f1a891a70candtopic[0]=0xf63780e752c6a54a94fc52715dbc5518a3b4c3c2833d301a204226548a2a8545\nexample:\n/api?module=logs&action=getLogs&fromBlock=379224&toBlock=latest&address=0x33990122638b9132ca29c723bdf037f1a891a70c&topic0=0xf63780e752c6a54a94fc52715dbc5518a3b4c3c2833d301a204226548a2a8545&apikey=YourApiKeyToken', 'title': 'GetEventLogsfromblocknumber379224tolatestBlock,wherelogaddress=0x33990122638b9132ca29c723bdf037f1a891a70candtopic[0]=0xf63780e752c6a54a94fc52715dbc5518a3b4c3c2833d301a204226548a2a8545', 'pieces': {'module': ['logs'], 'action': ['getLogs'], 'fromBlock': [], 'toBlock': ['latest'], 'address': [], 'topic0': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['fromBlock', 'address', 'topic0'], 'example': '/api?module=logs&action=getLogs&fromBlock=379224&toBlock=latest&address=0x33990122638b9132ca29c723bdf037f1a891a70c&topic0=0xf63780e752c6a54a94fc52715dbc5518a3b4c3c2833d301a204226548a2a8545&apikey=YourApiKeyToken'}, 'Th_blknum': {'description': 'eth_blockNumber\nexample:\n/api?module=proxy&action=eth_blockNumber&apikey=YourApiKeyToken', 'title': 'eth_blockNumber', 'pieces': {'module': ['proxy'], 'action': ['eth_blockNumber'], 'apikey': ['YourApiKeyToken']}, 'inputs': [], 'example': '/api?module=proxy&action=eth_blockNumber&apikey=YourApiKeyToken'}, 'Th_getblkbynum': {'description': 'eth_getBlockByNumber\nexample:\n/api?module=proxy&action=eth_getBlockByNumber&tag=0x10d4f&boolean=true&apikey=YourApiKeyToken', 'title': 'eth_getBlockByNumber', 'pieces': {'module': ['proxy'], 'action': ['eth_getBlockByNumber'], 'tag': [], 'boolean': ['true'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['tag'], 'example': '/api?module=proxy&action=eth_getBlockByNumber&tag=0x10d4f&boolean=true&apikey=YourApiKeyToken'}, 'Th_getunclebyblknum&index': {'description': 'eth_getUncleByBlockNumberAndIndex\nexample:\n/api?module=proxy&action=eth_getUncleByBlockNumberAndIndex&tag=0x210A9B&index=0x0&apikey=YourApiKeyToken', 'title': 'eth_getUncleByBlockNumberAndIndex', 'pieces': {'module': ['proxy'], 'action': ['eth_getUncleByBlockNumberAndIndex'], 'tag': [], 'index': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['tag', 'index'], 'example': '/api?module=proxy&action=eth_getUncleByBlockNumberAndIndex&tag=0x210A9B&index=0x0&apikey=YourApiKeyToken'}, 'Th_getblktxncountbynum': {'description': 'eth_getBlockTransactionCountByNumber\nexample:\n/api?module=proxy&action=eth_getBlockTransactionCountByNumber&tag=0x10FB78&apikey=YourApiKeyToken', 'title': 'eth_getBlockTransactionCountByNumber', 'pieces': {'module': ['proxy'], 'action': ['eth_getBlockTransactionCountByNumber'], 'tag': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['tag'], 'example': '/api?module=proxy&action=eth_getBlockTransactionCountByNumber&tag=0x10FB78&apikey=YourApiKeyToken'}, 'Th_gettxnbyhash': {'description': 'eth_getTransactionByHash\nexample:\n/api?module=proxy&action=eth_getTransactionByHash&txhash=0x1e2910a262b1008d0616a0beb24c1a491d78771baa54a33e66065e03b1f46bc1&apikey=YourApiKeyToken', 'title': 'eth_getTransactionByHash', 'pieces': {'module': ['proxy'], 'action': ['eth_getTransactionByHash'], 'txhash': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['txhash'], 'example': '/api?module=proxy&action=eth_getTransactionByHash&txhash=0x1e2910a262b1008d0616a0beb24c1a491d78771baa54a33e66065e03b1f46bc1&apikey=YourApiKeyToken'}, 'Th_gettxnbyblknum&index': {'description': 'eth_getTransactionByBlockNumberAndIndex\nexample:\n/api?module=proxy&action=eth_getTransactionByBlockNumberAndIndex&tag=0x10d4f&index=0x0&apikey=YourApiKeyToken', 'title': 'eth_getTransactionByBlockNumberAndIndex', 'pieces': {'module': ['proxy'], 'action': ['eth_getTransactionByBlockNumberAndIndex'], 'tag': [], 'index': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['tag', 'index'], 'example': '/api?module=proxy&action=eth_getTransactionByBlockNumberAndIndex&tag=0x10d4f&index=0x0&apikey=YourApiKeyToken'}, 'Th_gettxncount': {'description': 'eth_getTransactionCount\nexample:\n/api?module=proxy&action=eth_getTransactionCount&address=0x2910543af39aba0cd09dbb2d50200b3e800a63d2&tag=latest&apikey=YourApiKeyToken', 'title': 'eth_getTransactionCount', 'pieces': {'module': ['proxy'], 'action': ['eth_getTransactionCount'], 'address': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=proxy&action=eth_getTransactionCount&address=0x2910543af39aba0cd09dbb2d50200b3e800a63d2&tag=latest&apikey=YourApiKeyToken'}, 'Th_sendrawtxn': {'description': 'eth_sendRawTransaction\nexample:\n/api?module=proxy&action=eth_sendRawTransaction&hex=0xf904808000831cfde080&apikey=YourApiKeyToken', 'title': 'eth_sendRawTransaction', 'pieces': {'module': ['proxy'], 'action': ['eth_sendRawTransaction'], 'hex': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['hex'], 'example': '/api?module=proxy&action=eth_sendRawTransaction&hex=0xf904808000831cfde080&apikey=YourApiKeyToken'}, 'Th_gettxnreceipt': {'description': 'eth_getTransactionReceipt\nexample:\n/api?module=proxy&action=eth_getTransactionReceipt&txhash=0x1e2910a262b1008d0616a0beb24c1a491d78771baa54a33e66065e03b1f46bc1&apikey=YourApiKeyToken', 'title': 'eth_getTransactionReceipt', 'pieces': {'module': ['proxy'], 'action': ['eth_getTransactionReceipt'], 'txhash': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['txhash'], 'example': '/api?module=proxy&action=eth_getTransactionReceipt&txhash=0x1e2910a262b1008d0616a0beb24c1a491d78771baa54a33e66065e03b1f46bc1&apikey=YourApiKeyToken'}, 'Th_call': {'description': 'eth_call\nexample:\n/api?module=proxy&action=eth_call&to=0xAEEF46DB4855E25702F8237E8f403FddcaF931C0&data=0x70a08231000000000000000000000000e16359506c028e51f16be38986ec5746251e9724&tag=latest&apikey=YourApiKeyToken', 'title': 'eth_call', 'pieces': {'module': ['proxy'], 'action': ['eth_call'], 'to': [], 'data': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['to', 'data'], 'example': '/api?module=proxy&action=eth_call&to=0xAEEF46DB4855E25702F8237E8f403FddcaF931C0&data=0x70a08231000000000000000000000000e16359506c028e51f16be38986ec5746251e9724&tag=latest&apikey=YourApiKeyToken'}, 'Th_getcode': {'description': 'eth_getCode\nexample:\n/api?module=proxy&action=eth_getCode&address=0xf75e354c5edc8efed9b59ee9f67a80845ade7d0c&tag=latest&apikey=YourApiKeyToken', 'title': 'eth_getCode', 'pieces': {'module': ['proxy'], 'action': ['eth_getCode'], 'address': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=proxy&action=eth_getCode&address=0xf75e354c5edc8efed9b59ee9f67a80845ade7d0c&tag=latest&apikey=YourApiKeyToken'}, 'Th_getstorageat': {'description': 'eth_getStorageAt\nexample:\n/api?module=proxy&action=eth_getStorageAt&address=0x6e03d9cce9d60f3e9f2597e13cd4c54c55330cfd&position=0x0&tag=latest&apikey=YourApiKeyToken', 'title': 'eth_getStorageAt', 'pieces': {'module': ['proxy'], 'action': ['eth_getStorageAt'], 'address': [], 'position': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address', 'position'], 'example': '/api?module=proxy&action=eth_getStorageAt&address=0x6e03d9cce9d60f3e9f2597e13cd4c54c55330cfd&position=0x0&tag=latest&apikey=YourApiKeyToken'}, 'Th_gasprice': {'description': 'eth_gasPrice\nexample:\n/api?module=proxy&action=eth_gasPrice&apikey=YourApiKeyToken', 'title': 'eth_gasPrice', 'pieces': {'module': ['proxy'], 'action': ['eth_gasPrice'], 'apikey': ['YourApiKeyToken']}, 'inputs': [], 'example': '/api?module=proxy&action=eth_gasPrice&apikey=YourApiKeyToken'}, 'Th_estimategas': {'description': 'eth_estimateGas\nexample:\n/api?module=proxy&action=eth_estimateGas&to=0xf0160428a8552ac9bb7e050d90eeade4ddd52843&value=0xff22&gasPrice=0x051da038cc&gas=0xffffff&apikey=YourApiKeyToken', 'title': 'eth_estimateGas', 'pieces': {'module': ['proxy'], 'action': ['eth_estimateGas'], 'to': [], 'value': [], 'gasPrice': [], 'gas': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['to', 'value', 'gasPrice', 'gas'], 'example': '/api?module=proxy&action=eth_estimateGas&to=0xf0160428a8552ac9bb7e050d90eeade4ddd52843&value=0xff22&gasPrice=0x051da038cc&gas=0xffffff&apikey=YourApiKeyToken'}, 'Erc20-tknTotalsupplyByCntrctaddrs': {'description': 'GetERC20-TokenTotalSupplybyContractAddress\nexample:\n/api?module=stats&action=tokensupply&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&apikey=YourApiKeyToken', 'title': 'GetERC20-TokenTotalSupplybyContractAddress', 'pieces': {'module':
 ['stats'], 'action': ['tokensupply'], 'contractaddress': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['contractaddress'], 'example': '/api?module=stats&action=tokensupply&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&apikey=YourApiKeyToken'}, 'HistErc20-tknTotalsupplyByCntrctaddrs&Blkno': {'description': 'GetHistoricalERC20-TokenTotalSupplybyContractAddress&BlockNo\nexample:\n/api?module=stats&action=tokensupplyhistory&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&blockno=8000000&apikey=YourApiKeyToken', 'title': 'GetHistoricalERC20-TokenTotalSupplybyContractAddress&BlockNo', 'pieces': {'module': ['stats'], 'action': ['tokensupplyhistory'], 'contractaddress': [], 'blockno': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['contractaddress', 'blockno'], 'example': '/api?module=stats&action=tokensupplyhistory&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&blockno=8000000&apikey=YourApiKeyToken'}, 'I>[deprecated]</i>GetTknTotalsupplyByTknname': {'description': '<i>[Deprecated]</i>GetTokenTotalSupplybyTokenName(<divstyle=text-decoration:line-through>SupportedTokenNames:DGD,MKR,FirstBlood,HackerGold,ICONOMI,Pluton,REP,SNGLS</div>).Thishasfeaturebeendeprecated,insteadusetheApiabovetolookupanyERC20tokensupplybyitscontractaddress\nexample:\n/api?module=stats&action=tokensupply&tokenname=DGD&apikey=YourApiKeyToken', 'title': '<i>[Deprecated]</i>GetTokenTotalSupplybyTokenName(<divstyle=text-decoration:line-through>SupportedTokenNames:DGD,MKR,FirstBlood,HackerGold,ICONOMI,Pluton,REP,SNGLS</div>).Thishasfeaturebeendeprecated,insteadusetheApiabovetolookupanyERC20tokensupplybyitscontractaddress', 'pieces': {'module': ['stats'], 'action': ['tokensupply'], 'tokenname': ['DGD'], 'apikey': ['YourApiKeyToken']}, 'inputs': [], 'example': '/api?module=stats&action=tokensupply&tokenname=DGD&apikey=YourApiKeyToken'}, 'Erc20-tknAccountBalForTkncntrctaddrs': {'description': 'GetERC20-TokenAccountBalanceforTokenContractAddress\nexample:\n/api?module=account&action=tokenbalance&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&address=0xe04f27eb70e025b78871a2ad7eabe85e61212761&tag=latest&apikey=YourApiKeyToken', 'title': 'GetERC20-TokenAccountBalanceforTokenContractAddress', 'pieces': {'module': ['account'], 'action': ['tokenbalance'], 'contractaddress': [], 'address': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['contractaddress', 'address'], 'example': '/api?module=account&action=tokenbalance&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&address=0xe04f27eb70e025b78871a2ad7eabe85e61212761&tag=latest&apikey=YourApiKeyToken'}, 'HistErc20-tknAccountBalForTkncntrctaddrsByBlkno': {'description': 'GetHistoricalERC20-TokenAccountBalanceforTokenContractAddressbyBlockNo\nexample:\n/api?module=account&action=tokenbalancehistory&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&address=0xe04f27eb70e025b78871a2ad7eabe85e61212761&blockno=8000000&apikey=YourApiKeyToken', 'title': 'GetHistoricalERC20-TokenAccountBalanceforTokenContractAddressbyBlockNo', 'pieces': {'module': ['account'], 'action': ['tokenbalancehistory'], 'contractaddress': [], 'address': [], 'blockno': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['contractaddress', 'address', 'blockno'], 'example': '/api?module=account&action=tokenbalancehistory&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&address=0xe04f27eb70e025b78871a2ad7eabe85e61212761&blockno=8000000&apikey=YourApiKeyToken'}, 'I>[deprecated]</i>GetTknAccountBalByKnownTknname': {'description': '<i>[Deprecated]</i>GetTokenAccountBalancebyknownTokenName(<divstyle=text-decoration:line-through>SupportedTokenNames:DGD,MKR,FirstBlood,ICONOMI,Pluton,REP,SNGLS</div>).Thisfeaturehasbeendeprecated,insteadusetheApiabovetolookupanyERC20tokenbalancebyitscontractaddress\nexample:\n/api?module=account&action=tokenbalance&tokenname=DGD&address=0x4366ddc115d8cf213c564da36e64c8ebaa30cdbd&tag=latest&apikey=YourApiKeyToken', 'title': '<i>[Deprecated]</i>GetTokenAccountBalancebyknownTokenName(<divstyle=text-decoration:line-through>SupportedTokenNames:DGD,MKR,FirstBlood,ICONOMI,Pluton,REP,SNGLS</div>).Thisfeaturehasbeendeprecated,insteadusetheApiabovetolookupanyERC20tokenbalancebyitscontractaddress', 'pieces': {'module': ['account'], 'action': ['tokenbalance'], 'tokenname': ['DGD'], 'address': [], 'tag': ['latest'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['address'], 'example': '/api?module=account&action=tokenbalance&tokenname=DGD&address=0x4366ddc115d8cf213c564da36e64c8ebaa30cdbd&tag=latest&apikey=YourApiKeyToken'}, 'TknInfoByCntrctaddrs': {'description': 'GetTokenInfobyContractAddress\nexample:\n/api?module=token&action=tokeninfo&contractaddress=0x0e3a2a1f2146d86a604adc220b4967a898d7fe07&apikey=YourApiKeyToken', 'title': 'GetTokenInfobyContractAddress', 'pieces': {'module': ['token'], 'action': ['tokeninfo'], 'contractaddress': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['contractaddress'], 'example': '/api?module=token&action=tokeninfo&contractaddress=0x0e3a2a1f2146d86a604adc220b4967a898d7fe07&apikey=YourApiKeyToken'}, 'EstimationOfConfirmationTime': {'description': 'GetEstimationofConfirmationTime\nexample:\n/api?module=gastracker&action=gasestimate&gasprice=2000000000&apikey=YourApiKeyToken', 'title': 'GetEstimationofConfirmationTime', 'pieces': {'module': ['gastracker'], 'action': ['gasestimate'], 'gasprice': [], 'apikey': ['YourApiKeyToken']}, 'inputs': ['gasprice'], 'example': '/api?module=gastracker&action=gasestimate&gasprice=2000000000&apikey=YourApiKeyToken'}, 'GasOracle': {'description': 'GetGasOracle\nexample:\n/api?module=gastracker&action=gasoracle&apikey=YourApiKeyToken', 'title': 'GetGasOracle', 'pieces': {'module': ['gastracker'], 'action': ['gasoracle'], 'apikey': ['YourApiKeyToken']}, 'inputs': [], 'example': '/api?module=gastracker&action=gasoracle&apikey=YourApiKeyToken'}, 'TotalSupplyOfEth': {'description': 'GetTotalSupplyofEther\nexample:\n/api?module=stats&action=ethsupply&apikey=YourApiKeyToken', 'title': 'GetTotalSupplyofEther', 'pieces': {'module': ['stats'], 'action': ['ethsupply'], 'apikey': ['YourApiKeyToken']}, 'inputs': [], 'example': '/api?module=stats&action=ethsupply&apikey=YourApiKeyToken'}, 'EthLastPrice': {'description': 'GetETHERLastPrice\nexample:\n/api?module=stats&action=ethprice&apikey=YourApiKeyToken', 'title': 'GetETHERLastPrice', 'pieces': {'module': ['stats'], 'action': ['ethprice'], 'apikey': ['YourApiKeyToken']}, 'inputs': [], 'example': '/api?module=stats&action=ethprice&apikey=YourApiKeyToken'}, 'EtheumNodesSize': {'description': 'GetEthereumNodesSize\nexample:\n/api?module=stats&action=chainsize&startdate=2019-02-01&enddate=2019-02-28&clienttype=geth&syncmode=default&sort=asc&apikey=YourApiKeyToken', 'title': 'GetEthereumNodesSize', 'pieces': {'module': ['stats'], 'action': ['chainsize'], 'startdate': [], 'enddate': [], 'clienttype': ['geth'], 'syncmode': ['default'], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=chainsize&startdate=2019-02-01&enddate=2019-02-28&clienttype=geth&syncmode=default&sort=asc&apikey=YourApiKeyToken'}, 'EthHistPrice': {'description': 'GetETHERHistoricalPrice\nexample:\n/api?module=stats&action=ethdailyprice&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetETHERHistoricalPrice', 'pieces': {'module': ['stats'], 'action': ['ethdailyprice'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=ethdailyprice&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'EthHistDailyMarketCap': {'description': 'GetETHERHistoricalDailyMarketCap\nexample:\n/api?module=stats&action=ethdailymarketcap&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetETHERHistoricalDailyMarketCap', 'pieces': {'module': ['stats'], 'action': ['ethdailymarketcap'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=ethdailymarketcap&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyTxnCount': {'description': 'GetDailyTransactionCount\nexample:\n/api?module=stats&action=dailytx&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyTransactionCount', 'pieces': {'module': ['stats'], 'action': ['dailytx'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailytx&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyNewAddrsCount': {'description': 'GetDailyNewAddressCount\nexample:\n/api?module=stats&action=dailynewaddress&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyNewAddressCount', 'pieces': {'module': ['stats'], 'action': ['dailynewaddress'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailynewaddress&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyAverageBlkSize': {'description': 'GetDailyAverageBlockSize\nexample:\n/api?module=stats&action=dailyavgblocksize&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyAverageBlockSize', 'pieces': {'module': ['stats'], 'action': ['dailyavgblocksize'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyavgblocksize&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyAverageBlkTimeForBlkToBeIncludedInTheEtheumBlkchain': {'description':
 'GetDailyAverageBlockTimeforABlocktobeIncludedintheEthereumBlockchain\nexample:\n/api?module=stats&action=dailyavgblocktime&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyAverageBlockTimeforABlocktobeIncludedintheEthereumBlockchain', 'pieces': {'module': ['stats'], 'action': ['dailyavgblocktime'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyavgblocktime&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyAverageGasPriceUsed': {'description': 'GetDailyAverageGasPriceUsed\nexample:\n/api?module=stats&action=dailyavggasprice&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyAverageGasPriceUsed', 'pieces': {'module': ['stats'], 'action': ['dailyavggasprice'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyavggasprice&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyAverageGasLimit': {'description': 'GetDailyAverageGasLimit\nexample:\n/api?module=stats&action=dailyavggaslimit&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyAverageGasLimit', 'pieces': {'module': ['stats'], 'action': ['dailyavggaslimit'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyavggaslimit&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyTotalGasUsed': {'description': 'GetDailyTotalGasUsed\nexample:\n/api?module=stats&action=dailygasused&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyTotalGasUsed', 'pieces': {'module': ['stats'], 'action': ['dailygasused'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailygasused&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyBlkRew': {'description': 'GetDailyBlockRewards\nexample:\n/api?module=stats&action=dailyblockrewards&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyBlockRewards', 'pieces': {'module': ['stats'], 'action': ['dailyblockrewards'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyblockrewards&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyBlkCount&BlkRew': {'description': 'GetDailyBlockCountandBlockRewards\nexample:\n/api?module=stats&action=dailyblkcount&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyBlockCountandBlockRewards', 'pieces': {'module': ['stats'], 'action': ['dailyblkcount'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyblkcount&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyUncleBlkCount&UncleBlkRew': {'description': 'GetDailyUncleBlockCountandUncleBlockRewards\nexample:\n/api?module=stats&action=dailyuncleblkcount&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyUncleBlockCountandUncleBlockRewards', 'pieces': {'module': ['stats'], 'action': ['dailyuncleblkcount'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyuncleblkcount&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyAverageNetworkHashRate': {'description': 'GetDailyAverageNetworkHashRate\nexample:\n/api?module=stats&action=dailyavghashrate&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyAverageNetworkHashRate', 'pieces': {'module': ['stats'], 'action': ['dailyavghashrate'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyavghashrate&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyAverageNetworkDifficulty': {'description': 'GetDailyAverageNetworkDifficulty\nexample:\n/api?module=stats&action=dailyavgnetdifficulty&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyAverageNetworkDifficulty', 'pieces': {'module': ['stats'], 'action': ['dailyavgnetdifficulty'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyavgnetdifficulty&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyTotalTxnFee': {'description': 'GetDailyTotalTransactionFee\nexample:\n/api?module=stats&action=dailytxnfee&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyTotalTransactionFee', 'pieces': {'module': ['stats'], 'action': ['dailytxnfee'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailytxnfee&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyNetworkUtilization': {'description': 'GetDailyNetworkUtilization\nexample:\n/api?module=stats&action=dailynetutilization&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyNetworkUtilization', 'pieces': {'module': ['stats'], 'action': ['dailynetutilization'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailynetutilization&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}, 'DailyEnsRegistrationCount': {'description': 'GetDailyENSRegistrationCount\nexample:\n/api?module=stats&action=dailyensregister&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken', 'title': 'GetDailyENSRegistrationCount', 'pieces': {'module': ['stats'], 'action': ['dailyensregister'], 'startdate': [], 'enddate': [], 'sort': ['asc'], 'apikey': ['YourApiKeyToken']}, 'inputs': ['startdate', 'enddate'], 'example': '/api?module=stats&action=dailyensregister&startdate=2019-02-01&enddate=2019-02-28&sort=asc&apikey=YourApiKeyToken'}}
options = {'module': ['logs', 'account', 'contract', 'transaction', 'block', 'proxy', 'stats', 'token', 'gastracker'], 'action': ['getLogs', 'balance', 'balancehistory', 'balancemulti', 'txlist', 'txlistinternal', 'tokentx', 'tokennfttx', 'getminedblocks', 'getabi', 'getsourcecode', 'getstatus', 'gettxreceiptstatus', 'getblockreward', 'getblockcountdown', 'getblocknobytime', 'eth_blockNumber', 'eth_getBlockByNumber', 'eth_getUncleByBlockNumberAndIndex', 'eth_getBlockTransactionCountByNumber', 'eth_getTransactionByHash', 'eth_getTransactionByBlockNumberAndIndex', 'eth_getTransactionCount', 'eth_sendRawTransaction', 'eth_getTransactionReceipt', 'eth_call', 'eth_getCode', 'eth_getStorageAt', 'eth_gasPrice', 'eth_estimateGas', 'tokensupply', 'tokensupplyhistory', 'tokenbalance', 'tokenbalancehistory', 'tokeninfo', 'gasestimate', 'gasoracle', 'ethsupply', 'ethprice', 'chainsize', 'ethdailyprice', 'ethdailymarketcap', 'dailytx', 'dailynewaddress', 'dailyavgblocksize', 'dailyavgblocktime', 'dailyavggasprice', 'dailyavggaslimit', 'dailygasused', 'dailyblockrewards', 'dailyblkcount', 'dailyuncleblkcount', 'dailyavghashrate', 'dailyavgnetdifficulty', 'dailytxnfee', 'dailynetutilization', 'dailyensregister']}
inputs = ['fromBlock', 'toBlock', 'topic0', 'topic1', 'page', 'offset', 'address', 'blockno', 'startblock', 'endblock', 'txhash', 'timestamp', 'tag', 'index', 'hex', 'to', 'data', 'position', 'value', 'gasPrice', 'gas', 'contractaddress', 'gasprice', 'startdate', 'enddate','apikey']
def get_api_key(scanner,env_key=None):
    if env_key:
        return get_env_value(key = env_key,deep_scan=True)
    for network in [scanner,'etherscan']:
        env_value = get_env_value(key = f"{network}_api",deep_scan=True)
        if env_value:
            return env_value
def generate_api_variables(generated_url):
    js_all={}
    for each in generated_url.split('?')[-1].split('&'):
        api_key = each.split('=')[0]
        api_val = each.split('=')[1]
        if api_key not in js_all:
            js_all[api_key]= []
        if api_val not in js_all[api_key]:
            js_all[api_key].append(api_val)
    return js_all
def get_generated_url_output(scanner='etherscan.io',selected_api='CntrctSrcCodeForVerifiedCntrctSrcCodes',env_key=None,fromBlock=None,
                             toBlock=None,
                             topic0=None,
                             topic1=None,
                             page=None,
                             offset=None,
                             address=None,
                             contractaddress=None,
                             blockno=None,
                             startblock=None,
                             endblock=None,
                             txhash=None,
                             timestamp=None,
                             tag=None,
                             index=None,
                             hex=None,
                             to=None,
                             data=None,
                             position=None,
                             value=None,
                             gasPrice=None,
                             gas=None,
                             gasprice=None,
                             startdate=None,
                             enddate=None,
                             apikey=None):
    apikey=get_api_key(scanner,env_key=env_key)
    generated_url = apiCallDesc[selected_api]['example']
    values = generate_api_variables(generated_url=generated_url)
    if scanner:
        text=f"https://{('api.' if 'api' != scanner[:len('api')] else '')}{scanner}/api?module={values.get('module')[0]}&action={values.get('action')[0]}"
        js_all_keys = list(values.keys())
        js_all_keys = js_all_keys[2:]
        each_log = {}
        key_values = [fromBlock, toBlock, topic0, topic1, page, offset, address, blockno, startblock, endblock, txhash, timestamp, tag, index, hex, to, data, position, value, gasPrice, gas,contractaddress, gasprice, startdate, enddate,apikey]
        for i,key in enumerate(inputs):
            values[key] = key_values[i]
        for i,each in enumerate(js_all_keys):
            if each not in each_log:
                each_log[each]=0
            value=values.get(each,'')
            if value and isinstance(value,list):
                if len(value) > each_log[each]:
                    value = value[each_log[each]]
            text+="&"+f"{each}"+"="+f"{value}"
            each_log[each]+=1
        text = text.split('apikey=')[0]+f"apikey={apikey}"
        # Here's where you would combine the base URL with the user's inputs 
        # to generate the full API URL. I'm just simulating this step.
        return text

def checksum(address: str=None,rpc=None):
    """
    Attempt to convert the address to a checksum address.

    :param address: Ethereum address to convert.
    :return: Checksum Ethereum address.
    :raises ValueError: If the address is invalid.
    """
    if address == None:
        return
    checkSumRefference = create_and_read_json(file_path=get_check_sum_path(),data={})
    if isinstance(address,str):
        
        refferenced_address = checkSumRefference.get(address.lower())
        if refferenced_address:
            return refferenced_address
         
        checked_address = w3_mgr.retrieve_w3(address,rpc=rpc).to_checksum_address(address,)
        checkSumRefference[address.lower()]=checked_address
        safe_dump_to_file(file_path=get_check_sum_path(),data=checkSumRefference)
        return checked_address
def read_json(file_path):
    try:
        with open(file_path, 'r') as file:
            return json.load(file)
    except FileNotFoundError:
        return None
    except json.JSONDecodeError:
        raise ValueError(f"Invalid JSON in {file_path}")

def write_json(file_path, data):
    with open(file_path, 'w') as file:
        json.dump(data, file, indent=4)

def get_directory_path():
    return os.path.dirname(os.path.abspath(__file__))

def get_data_directory():
    return os.path.join(get_directory_path(), 'data')

def get_check_sum_path():
    return os.path.join(get_data_directory(), 'check_sums.json')

def get_source_code_directory():
    return os.path.join(get_data_directory(), 'source_codes')

def create_address_directory(address):
    return os.path.join(get_source_code_directory(),get_normalized_address(address))

def create_network_name_directory(address,network_name):
    return os.path.join(create_address_directory(address),get_normalized_file_name(network_name))

def get_response(response):
    # Improved error handling and response parsing
    try:
        if response.status_code == 200:
            return response.json()
        else:
            return response.text
    except Exception as e:
        print(f"Error processing response: {e}")
        return None
def strip_web(url):
    if isinstance(url,dict):
        url = url.get('url',url)        
    if url:
        if url.startswith("http://"):
            url = url.replace("http://", '', 1)
        elif url.startswith("https://"):
            url = url.replace("https://", '', 1)
        url = url.split('/')[0]
        return url
def make_request(url):
        response = get_response(requests.get(url))
        return response
def get_normalized_address(address):
    return address.lower()
def get_normalized_file_name(name):
    return name.lower().replace(' ','_')
def get_new_rpc_bridge(rpc_js=None):
    return RPCBridge(rpc_js=rpc_js)

# Simplified to focus on checking source code existence
def check_source_code(address):
    contract_path = create_address_directory(address)
    if os.path.isdir(contract_path):
        rpc_name = w3_mgr.retrieve_rpc_js(address).get('name')
        files = [rpc_name]+os.listdir(contract_path)
        for file in files:
            if file:
                path = os.path.join(contract_path,file)
                source_code_path = os.path.join(path,'source_code.json')
                if os.path.isfile(source_code_path):
                    return source_code_path
    
    

def save_source_code(rpc_js, source_code, address):
    # Simplified saving process
    address_dir = os.path.join(get_source_code_directory(), get_normalized_address(address))
    network_dir = os.path.join(address_dir,get_normalized_file_name(rpc_js['name']))
    rpc_path = os.path.join(network_dir, 'rpc_data.json')
    source_code_path = os.path.join(network_dir, 'source_code.json')
    os.makedirs(address_dir,exist_ok=True)
    os.makedirs(network_dir,exist_ok=True)
    write_json(file_path=rpc_path,data=rpc_js)
    write_json(file_path=source_code_path,data=source_code)
    return network_dir

def get_source_code_path(address, rpc_js=None):
    
    source_code_path = check_source_code(address)
    if source_code_path:
        return os.path.dirname(source_code_path)
    derive_network(address=address)
    source_code_path = check_source_code(address)
    if source_code_path:
        return os.path.dirname(source_code_path)
    

def get_response(response):
    get_response,count = None,0
    while get_response == None and count <3:
        try:
            if count ==2:
                get_response = response.text
            if count ==1:
                get_response = response.json()
            if count == 0:
                status_code = response.status_code
        except Exception as e:
            print(e)
        count+=1
    response = get_response or response
    if isinstance(response,dict):
        response = response.get('result',response)
    return safe_json_loads(response)
def make_request(url):

    return get_response(requests.get(url))
    

def derive_network(address=None,multiple=False,initial_network='ethereum'):
    networks_found = []
    
    rpc_mgr_reference = get_new_rpc_bridge(rpc_js=initial_network)
    rpc_list,common_chains = get_default_rpc_list(),rpc_mgr_reference.get_common_chains()
    initial_network = initial_network or rpc_mgr_reference.name or common_chains[0]
    
    address = checksum(address=address)
    source_code = check_source_code(address)
    if source_code:
        return {"network":rpc_mgr_reference.rpc_js,"contract_data":read_json(source_code)}
    network_list = make_list(initial_network)+common_chains+rpc_list

    for list_itteration,rpc_js in enumerate(network_list):
        found = False
        if rpc_js:
            rpc_js=rpc_js
            if list_itteration != 2:
                rpc_js = [rpc_js]
            rpc_mgr = RPCBridge(rpc_js=rpc_js)
            

            url=get_generated_url_output(scanner=rpc_mgr.scanner,address=address)
            response =make_request(url)
            if response and isinstance(response,list):
                if response[0] and isinstance(response[0],dict):
                    if "ContractName" in response[0].keys() and response[0].get("ContractName") not in [None,'',' ']:
                        found = True 
            if found:
               
                if multiple:
                    save_source_code(rpc_mgr.rpc_js,response,address)
                    networks_found.append({"network":rpc_mgr.rpc_js,"contract_data":response})
                else:
                    save_source_code(rpc_mgr_reference.rpc_js,  response, address)
                    return {"network":rpc_mgr_reference.rpc_js,"contract_data":response}
                     
        
            rpc_list = remove_from_list(rpc_list,'name',rpc_mgr_reference.rpc_js['name'])
    return networks_found




from .rpc_functions import *
import json

class RPCBridge: 
    """
    RPCBridge class manages RPC parameters for a blockchain.
    """

    def __init__(self, rpc_js: dict = None, testnet=False, rpc_list=None, rpc_gui=False):
        """
        Initializes the RPCBridge instance with RPC parameters.
        """
        self.script_path = os.path.abspath(__file__)
        self.directory_path = os.path.dirname(self.script_path)
        self.rpc_list = rpc_list or get_default_rpc_list()
        self.testnet_rpcs, self.non_testnet_rpcs = separate_testnet_rpcs(self.rpc_list)
        self.non_testnet_values = get_total_values_list(self.non_testnet_rpcs)
        self.testnet_values = get_total_values_list(self.testnet_rpcs)
        self.rpc_values = {"False":[self.non_testnet_values,self.non_testnet_rpcs],"True":[self.testnet_values,self.testnet_rpcs]}
        self.common_chains = get_common_chains()
        self.rpc_js = rpc_js or get_default_rpc()
        self.symbol = self.rpc_js.get('chain', '')
        self.name = self.rpc_js.get('name', '')
        self.explorers = self.rpc_js.get('explorers', [])
        self.explorer = self.explorers[0] if self.explorers else None
        self.rpcs = get_rpc_urls(self.rpc_js)
        self.rpc= self.rpcs[0] if self.rpcs else None
        self.chainId = int(self.rpc_js['chainId']) if is_number(self.rpc_js['chainId']) else None
        self.scanner = strip_web(self.explorer['url'])
        self.w3 = Web3(Web3.HTTPProvider(self.rpc)) if self.rpc else None
        self.rpc_js['user_settings'] = {'symbol':self.symbol,
                                        'name':self.name,
                                        'explorers':self.explorers,
                                        'explorer':self.explorer,
                                        'rpcs':self.rpcs,
                                        'rpc':self.rpc,
                                        'chainId':self.chainId,
                                        'scanner':self.scanner,
                                        'w3':self.w3}
    def update_rpc_js(self, new_rpc_js: dict):
        """
        Directly updates the RPC JavaScript object with new settings.
        """
 
        if isinstance(new_rpc_js, dict):
            self.rpc_js = new_rpc_js
            self.setup_rpc_attributes()
            return self.rpc_js
    def setup_rpc_attributes(self):
        """
        Sets up various RPC attributes based on the current rpc_js.
        """
        # Extract values from rpc_js and set them to the instance attributes
        self.symbol = self.rpc_js.get('chain', '')
        self.name = self.rpc_js.get('name', '')
        self.explorers = self.rpc_js.get('explorers', [])
        self.explorer = self.explorers[0] if self.explorers else None
        self.rpcs = get_rpc_urls(self.rpc_js)
        self.rpc= self.rpcs[0] if self.rpcs else None
        self.chainId = int(self.rpc_js['chainId']) if is_number(self.rpc_js['chainId']) else None
        self.scanner = strip_web(self.explorer['url'])
        self.w3 = Web3(Web3.HTTPProvider(self.rpc)) if self.rpc else None
        self.rpc_js['user_settings'] = {'symbol':self.symbol,
                                        'name':self.name,
                                        'explorers':self.explorers,
                                        'explorer':self.explorer,
                                        'rpcs':self.rpcs,
                                        'rpc':self.rpc,
                                        'chainId':self.chainId,
                                        'scanner':self.scanner,
                                        'w3':self.w3}
    def update_settings(self, new_settings):
        """
        Updates the RPCBridge instance settings based on the provided new settings.

        Parameters:
        new_settings (dict): A dictionary containing the new settings.
        """
        # Update relevant attributes based on new settings
        self.rpc_js = new_settings.get('rpc_js', self.rpc_js)
        self.rpc = self.get_rpc_url(self.rpc_js)  # Assuming a method to extract the main RPC URL
        self.w3 = Web3(Web3.HTTPProvider(self.rpc)) if self.rpc else None
        # Add other necessary updates based on new settings

    # Helper method to extract the main RPC URL from rpc_js
    def get_rpc_url(self, rpc_js):
        return rpc_js['rpc'][0] if 'rpc' in rpc_js and len(rpc_js['rpc']) > 0 else None


class RPCGUIManager:
    def __init__(self, rpc_mgr=None,rpc_list:list=None,rpc_js=None,rpc_gui=True,gui_window=True):
        self.rpc_list = rpc_list
        self.rpc_mgr = rpc_mgr
        if self.rpc_mgr == None:
            self.rpc_mgr = RPCBridge(rpc_js=rpc_js,rpc_list=rpc_list)
        
        self.rpc_js = rpc_js or self.rpc_mgr.rpc_js
        self.user_selection =  self.rpc_js
        self.original_query=''
        self.rpc_list = self.rpc_mgr.rpc_list
        self.relevant_list = self.rpc_list
        self.section="RPCmgr"
        self.network_names = list({item.get('name') for item in self.rpc_list})
        self.rpc_key_value_js = {"name":"-NETWORK_NAME-",'Network':"-NETWORK-",'chain':"-SYMBOL-",'chainId':"-CHAINID-",'rpc':"-RPC-",'explorer':"-BLOCK_EXPLORER-"}
        self.all_keys = list(self.rpc_key_value_js.values())+['-SEARCH-',"-OK_RPC-","-EXIT_RPC-"]
        self.network_names.sort()
        self.new_network_names=self.network_names
        self.window_mgr = AbstractWindowManager()
        if gui_window:
            self.run_gui_window()
    def run_gui_window(self):
            self.window_name = self.window_mgr.add_window(title='RPC Selector',
                                                          layout=self.get_rpc_layout(),
                                                          close_events=[text_to_key("-OK_RPC-",section=self.section),
                                                                        text_to_key("-EXIT_RPC-",section=self.section)],
                                                          event_handlers=[self.rpc_win_while],
                                                          suppress_raise_key_errors=False,
                                                          suppress_error_popups=False,
                                                          suppress_key_guessing=False,
                                                          finalize=True)
            self.window=self.window_mgr.get_window()
            self.window_mgr.while_window(window_name=self.window_name)
            self.values = self.window_mgr.search_closed_windows(window_name=self.window_name,
                                                                window=self.window)['values']
            self.rpc_values = self.get_rpc_values(self.values)
    def get_rpc_layout(self,section=None):
        relevant_list=self.rpc_mgr.rpc_list
        rpc_js = self.rpc_mgr.rpc_js
        names = [item.get('name') for item in relevant_list]
        name = names[0]
        names,name,rpc_js=self.get_rpc_js_vars(relevant_list,names=names,name=name,rpc_js=rpc_js)
        name,networks,network,chain_ids,rpcs,rpc,explorers,explorer,symbol=self.get_static_variables(rpc_js)
        layout = [
            [sg.Text('SEARCH'), sg.Input('', key=text_to_key('SEARCH',section=self.section),size=(20,1), enable_events=True)],
            [sg.Text('Network Name:'), sg.Combo(self.network_names,default_value=name, key=text_to_key('NETWORK_NAME',section=self.section), enable_events=True)],
            [sg.Text('Network:'), sg.Combo(networks,default_value=network,key=text_to_key('NETWORK',section=self.section),size=(20,1), enable_events=True)],
            [sg.Text('RPC:'), sg.Combo(rpcs,default_value=rpc, key=text_to_key('RPC',section=self.section),size=(20,1), enable_events=True)],
            [sg.Text('ChainID:'), sg.InputText(chain_ids,key=text_to_key('CHAINID',section=self.section),size=(20,1), disabled=True)],  # Make this an InputText to display ChainID
            [sg.Text('Block Explorer:'), sg.Combo(explorers,default_value=explorer, key=text_to_key('BLOCK_EXPLORER',section=self.section),size=(20,1), enable_events=True)],
            [sg.Text('Symbol:'), sg.InputText(symbol,key=text_to_key('SYMBOL',section=self.section),size=(20,1), disabled=True)]  # Make this an InputText to display Symbol
        ]
        layout.append(create_row_of_buttons({"button_text":"OK","enable_event":True,"key":text_to_key("OK_RPC",section=self.section)},
                                            {"button_text":"Show","enable_event":True,"key":text_to_key("SHOW_RPC",section=self.section)},
                                            {"button_text":"Reset","enable_event":True,"key":text_to_key("RESET_RPC",section=self.section)},
                                            {"button_text":"Exit","enable_event":True,"key":text_to_key("EXIT_RPC",section=self.section)}))
        return layout
    def get_testnet_or_mainnet(self,string):
        return 'Testnet' if 'Testnet'.lower() in str(string).lower() else 'Mainnet'
    
    def get_key_from_value(self,value):
            """
            Fetches the key for a given value from the `get_rpc_js()` mapping.
            
            Parameters:
            - value: The value for which the key needs to be found.
            
            Returns:
            - The key corresponding to the value.
            """
            for key,key_value in self.rpc_key_value_js.items():
                if key_value == value:
                    return key
    def get_static_variables(self,rpc_js):
        name,networks,network,chain_ids,rpcs,rpc,explorers,explorer,symbol='',[],'','',[],'',[],'',''
        if rpc_js:
            name = rpc_js['name']
            network=self.get_testnet_or_mainnet(name)
            networks=[network]
            chain_ids = rpc_js['chainId']
            rpcs = self.rpc_mgr.get_rpc_url(rpc_js)
            if rpcs and isinstance(rpcs,list):
                rpc=rpcs[0]
            explorers = get_explorers(rpc_js)
            if explorers:
                explorer=explorers[0]
            symbol = rpc_js['chain']
        return name,networks,network,chain_ids,rpcs,rpc,explorers,explorer,symbol
    def get_rpc_js_vars(self,relevant_list,names=None,rpc_js=None,name=None):
        relevant_list= make_list(relevant_list or  self.rpc_mgr.rpc_list)
        
        if names ==None:
            names = [item['name'] for item in relevant_list]
        if name == None:
            name = names[0]
        if rpc_js == None:
            rpc_js = [item for item in relevant_list if item['name'] == name][0]
        return names,name,rpc_js
    def update_network_name(self,relevant_list,names=None,rpc_js=None,name=None):
        names,name,rpc_js=self.get_rpc_js_vars(relevant_list,names=names,rpc_js=rpc_js,name=name)
        self.window[self.script_event_js['-NETWORK_NAME-']].update(values=names,value=name)
        name,networks,network,chain_ids,rpcs,rpc,explorers,explorer,symbol=self.get_static_variables(rpc_js)
        self.window[self.script_event_js['-NETWORK-']].update(network)
        self.window[self.script_event_js['-RPC-']].update(values=rpcs,value=rpc)
        self.window[self.script_event_js['-CHAINID-']].update(value=chain_ids)
        self.window[self.script_event_js['-BLOCK_EXPLORER-']].update(values=explorers,value=explorer)
        self.window[self.script_event_js['-SYMBOL-']].update(value=symbol)

    def get_back_values(self,num_list,values):
        rpc_values = []
        for num in num_list:
            rpc_values.append(values[num])
        return rpc_values
    def rpc_win_while(self,event,values,window):
        self.event,self.values,self.window=event,values,window
        self.script_event_js = get_event_key_js(self.event,key_list=self.all_keys)
        value = None
        if self.script_event_js['found'] == '-SEARCH-':
            query = values[self.script_event_js['-SEARCH-']]
            if query == '' or len(query) < len(self.original_query):
                self.relevant_list = self.rpc_mgr.rpc_list
            else:
                self.relevant_list = [item for item in self.rpc_mgr.rpc_list if query.lower() in item['name'].lower()]
            name = self.values[self.script_event_js['-NETWORK_NAME-']]
            self.update_network_name(name=name,relevant_list=self.relevant_list)
                    #self.update_network_name(values=values,value=value,relevant_list=self.relevant_list)
        if self.script_event_js['found'] == '-OK_RPC-':

            self.apply_selected_rpc_settings(self.get_rpc_values(values))
        if self.script_event_js['found'] in ['-NETWORK_NAME-','-NETWORK-']:
            name = self.values[self.script_event_js['-NETWORK_NAME-']]
            relevant_list = self.rpc_mgr.rpc_list
            if self.script_event_js['found'] == '-NETWORK_NAME-':
                rpc_js = [item for item in self.rpc_mgr.rpc_list if item['name'] ==name][0] 
            else:
                network = self.get_testnet_or_mainnet(name)
                rpc_js = [item for item in self.rpc_mgr.rpc_list if item.get('name') ==name and self.get_testnet_or_mainnet(item.get('name')) == network][0] 
            selected_network = self.values[self.script_event_js['-NETWORK-']]
            self.update_network_name(rpc_js=rpc_js,names=self.new_network_names,name=name,relevant_list=relevant_list)
      
    def get_rpc_values(self,values) -> dict or None:
        rpc=[]
        section_keys = [self.script_event_js['-NETWORK_NAME-'],
                        self.script_event_js['-SYMBOL-'],
                        self.script_event_js['-CHAINID-'],
                        self.script_event_js['-NETWORK-'],
                        self.script_event_js['-BLOCK_EXPLORER-'],
                        self.script_event_js['-RPC-']]
        rpc_js = [item for item in self.rpc_mgr.rpc_list if item['name'] == values[section_keys[0]]][0]
        rpc_js['user_settings']={}
        for original_key in ['-NETWORK_NAME-','-SYMBOL-','-CHAINID-','-NETWORK-','-BLOCK_EXPLORER-','-RPC-']:
            event_js_key = self.script_event_js[original_key]
            rpc_key = self.get_key_from_value(original_key)
            value = values[event_js_key]
            rpc_js['user_settings'][rpc_key]=value
        
        return rpc_js
    def apply_selected_rpc_settings(self, gui_values):
        """
        Applies the selected RPC settings to the RPC manager based on the GUI values.
        """
        # Extract the necessary information from gui_values
        selected_rpc_js = self.extract_rpc_settings_from_gui(gui_values)
        if selected_rpc_js:
            # Update the RPC manager with the selected settings
            self.rpc_mgr.update_rpc_js(selected_rpc_js)
            # Reinitialize the Web3 instance with the new RPC
            self.rpc_mgr.w3 = Web3(Web3.HTTPProvider(self.rpc_mgr.rpc))

    def extract_rpc_settings_from_gui(self, gui_values):
        # Logic to extract and format the settings from the GUI values
        # Return the formatted settings
        self.rpc_mgr.rpc_js = self.user_selection = gui_values
        return self.rpc_mgr.rpc_js

RPCBridge()
